#pragma once
#include "GLRenderer.h"

struct REGION
{
  int noofpts, maxpts;
  POINTDATA** points;
};

#define REGION_SCALE  10.0f  // the larger the value the smaller the grid
#define REGION_SIZE   (1.0f/REGION_SCALE)

class CloudData
{
public:
  CloudData();
  ~CloudData();

  void SetNoOfBuffers(const int n);

  void ClearPointData();
  void AddPointData(POINTDATA* point);
  void AddPointData(POINTDATA* points, const int noofpts);
  void Prune(POINTDATA* wlddata, int& noofwldpts, const float* minxyz, const float* maxxyz, const int iteration);
  void CheckPrune(POINTDATA* wlddata, int& noofwldpts, POINTDATA* wldpt, const int iteration);

  bool CheckConsistency();
  POINTDATA** GetPoints(const float x, const float y, const float z, const float tolerance, int& noofpts, const int bufferno);
  POINTDATA** GetPoints(float* min, float* max, int& noofpts, const int bufferno, const bool reducewindow);
  POINTDATA** GetPoints(float* min, float* max, const double* inverse, const float* f_min, const float* f_max, int& noofpts, const int bufferno, const bool reducewindow);
  POINTDATA** GetPoints(const int noofcentres, const float* centres, const float radius, int& noofpts, const int bufferno);
//  POINTDATA** GetPoints(float* min, float* max, float* inverse, float* min_l, float* max_l, int& noofpts, const int bufferno, const bool reducewindow);
//  POINTDATA** GetPoints(const float x1, const float y1, const float z1, const float x2, const float y2, const float z2, int& noofpts, const int bufferno, const bool reducewindow);

private:
  int width, height, depth;
  int starts[3];
  REGION** regions;

  int noofbuffers;
  int* templistsize;
  POINTDATA*** templist;

  void ResizeRegionsX(int amount);
  void ResizeRegionsY(int amount);
  void ResizeRegionsZ(int amount);

  POINTDATA** GetPointsAux(int* minxyz, int* maxxyz, int& noofpts, const int bufferno);
  void GetIndicies(int* indicies, const float x, const float y, const float z);
  void ClampIndicies(int* indicies);
};

inline void CloudData::AddPointData(POINTDATA* points, const int noofpts)
{
  for (int i=0; i<noofpts; i++)
    AddPointData(points+i);
}

inline void CloudData::ClampIndicies(int* indicies)
{
  if (indicies[0]<0)
    indicies[0]=0;
  else if (indicies[0]>=width)
    indicies[0]=width-1;
  
  if (indicies[1]<0)
    indicies[1]=0;
  else if (indicies[1]>=height)
    indicies[1]=height-1;
  
  if (indicies[2]<0)
    indicies[2]=0;
  else if (indicies[2]>=depth)
    indicies[2]=depth-1;
}
