#pragma once
#include <Windows.h>

#define GLREDRAW        WM_USER+1

struct POINTDATA
{
  bool used, paired;
  float xyz[3], colour[3], nxyz[3];

  DWORD lastseen, count;

  POINTDATA* matchpoint;
  float matchdist, matchweight;
};

struct CAMERAINFO
{
  int width, height;
  float neardist, fardist;
};

struct CAMERATRACK
{
  double txyz[3], rxyz[3]; // rotations are expressed in radians
  double sins[3], coss[3];

//  float rotation[16], transform[16];  // down then across; [1]=first column, second row
};


class GLRenderer
{
public:
  GLRenderer();
  ~GLRenderer();

//  void Init(HWND hwnd);
  void Init(HWND parent, HINSTANCE hinstance, int x, int y, int width, int height, bool preview);
  void OnSize(UINT nType, int cx, int cy);
  void SetWindowPos(int x, int y, int width, int height);
  void GetWindowPos(int& x, int& y, int& width, int&height);
  int GetWindowPosX();
  void Draw();
  bool HitTest(int x, int y);
  void ResetCamera(bool redraw);

  void OnLButtonDown(UINT nType, int x, int y);
  void OnMouseMove(UINT nType, int x, int y);
  void OnLButtonUp(UINT nType, int x, int y);
  void OnMouseWheel(WPARAM wParam, int x, int y);

  void SetCameraTrack(CAMERATRACK* track);
  void SetCameraNearFar(float n, float f, int width, int height);
  void SetPointData(POINTDATA* pointdata, const int noofpoints, const bool redraw);

  HWND parent, hwnd;

private:
  HINSTANCE hinstance;
  CRITICAL_SECTION lock;

  bool preview;
  int x, y, width, height;
  
  float rotation[3];
  float eye[3], centre[3], up[3];
      
  bool mousedown;
  POINT lastpoint;

  unsigned int cube;
  
  int noofpoints;
  POINTDATA* pointdata;
//  bool drawcamerainfo;
  CAMERATRACK* cameratrack;
  CAMERAINFO camerainfo;

  void SetupLighting();
  void DrawPointCloud();
  void DrawCameraInfo();
  void CalcRotationMatrix(float mat[16]);
  void CalcLookAt(bool fromeye);

  void CreateCube();

  void InitAux();

  friend LRESULT CALLBACK GLWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
  friend DWORD WINAPI CreateGLRenderer(LPVOID lpParameter);
};

inline bool GLRenderer::HitTest(int x, int y)
{
  return (x>=this->x && x<=(this->x+width) && y>=this->y && y<=(this->y+height));
}

inline void GLRenderer::SetCameraNearFar(float n, float f, int width, int height)
{
  camerainfo.neardist=n;
  camerainfo.fardist=f;
  camerainfo.width=width;
  camerainfo.height=height;
}

inline int GLRenderer::GetWindowPosX()
{
  return x;
}

inline void GLRenderer::SetCameraTrack(CAMERATRACK* track)
{
  cameratrack=track;
}
