#include "MainWnd.h"
#include <assert.h>
#include <time.h>

MainWnd* mainwnd;
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
DWORD WINAPI ListenThread(LPVOID lpParameter);
DWORD WINAPI JacobianThread(LPVOID lpParameter);
DWORD WINAPI ICPErrorThread(LPVOID lpParameter);

DWORD WINAPI PlaybackThread(LPVOID lpParameter);

//#define PLAYBACK_TIMER    1001

#define UNKNOWN_COL       0x00000080
#define TOO_CLOSE_COL     0x00008000
#define TOO_FAR_COL       0x00800000

#define RECORD_FILE         L"Record.dat"
#define SETTINGS_FILE       L"Settings.dat"
#define MIN_MAX_FILE        L"DepthMinMaxs.csv"
#define AVERAGE_FILE        L"DepthAverages.csv"
#define DEPTH_FILE          L"DepthInfo.csv"
#define MATCH_FILE          L"MatchInfo.dat"
#define MODEL_OUTPUT        L"SceneExport-"
#define POLYGON_TEMP_FILE   L"PointsToPolys-"
#define SCREEN_GRAB_FOLDER  L"screens\\"

MainWnd::MainWnd(HINSTANCE hInstance) : hinstance(hInstance), sensor(NULL), colstream(NULL), depthstream(NULL), tracking(false), grabdata(false), recording(false), recorder(NULL), playing(false), noofthreads(0), alerts(NULL), jacobthreads(NULL), jacobthreadinfo(NULL), icpthreadinfo(NULL)
{
	WNDCLASSEX classname;
  classname.cbSize=sizeof(WNDCLASSEX); 
  classname.style=CS_HREDRAW|CS_VREDRAW;
  classname.lpfnWndProc=(WNDPROC) WndProc;
  classname.cbClsExtra=0;
  classname.cbWndExtra=0;
  classname.hInstance=hInstance;
  classname.hIcon=LoadIcon(NULL, IDI_WINLOGO);
  classname.hCursor=LoadCursor(NULL, IDC_ARROW);
  classname.hbrBackground=::GetSysColorBrush(COLOR_3DFACE);
  classname.lpszMenuName=NULL;
  classname.lpszClassName=L"KinectUp";
  classname.hIconSm=LoadIcon(NULL, IDI_WINLOGO);
  RegisterClassEx(&classname);

  memset(&colprev, 0, sizeof(PREVIEWINFO));
  memset(&depthprev, 0, sizeof(PREVIEWINFO));
  memset(&cameratrack, 0, sizeof(CAMERATRACK));

  exportdata[0]=exportdata[1]=false;

  InitTrackingMemory();

  SetNoOfThreads(8);

  pdupdate.SetCloudData(&clouddata);

//  TestCodeV2();

  int width=900;
  int height=720;
  int offx=(::GetSystemMetrics(SM_CXSCREEN)-width)/2;
  int offy=(::GetSystemMetrics(SM_CYSCREEN)-height)/2;

  for (int i=0; i<9; i++)
    calibration[i]=0.0f;
  calibration[0]=calibration[4]=calibration[8]=1.0f;

  trackedframes=0;
  lastgrab=0;
  lastpruned=0;

  maxwldpts=noofwldpts=0;
  worlddata=NULL;
/*
  depthseen=(bool*) malloc(sizeof(bool)*DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT);
  depthmins=(float*) malloc(sizeof(float)*DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT);
  depthmaxs=(float*) malloc(sizeof(float)*DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT);
  depthcounts=(int*) malloc(sizeof(int)*DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT);
  depthavgs=(float*) malloc(sizeof(float)*DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT);
  */
  LoadSettings();

  ::mainwnd=this;
  hwnd=CreateWindowEx(NULL, L"KinectUp", L"Kinecting up the Past", WS_OVERLAPPEDWINDOW|WS_VISIBLE, offx, offy, width, height, NULL, NULL, hInstance, NULL);

  MSG msg;
  do
  {
    if (GetMessage(&msg, NULL, 0, 0))
    {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
  } while (msg.message!=WM_QUIT);
/*
  free(depthseen);
  free(depthmins);
  free(depthmaxs);
  free(depthcounts);
  free(depthavgs);
  */
  KillThreads();
  FreeTrackingMemory();
  SaveSettings();
  Sleep(1000);
}

MainWnd::~MainWnd()
{
  kinectalive=false;
  SetEvent(frameevents[0]);
  Sleep(200);

  if (sensor)
  {
    sensor->NuiShutdown();
    sensor->Release();
  }

  free(pointdata);
  free(worlddata);

  ::DeleteObject(colprev.hbmpprev);
  ::DeleteObject(depthprev.hbmpprev);
  ::DeleteDC(hdctmp);

  CloseHandle(frameevents[0]);
  CloseHandle(frameevents[1]);

  KillThreads();
  free(jacobthreadinfo);
  free(icpthreadinfo);
}

DWORD WINAPI ListenThread(LPVOID lpParameter)
{
  ((MainWnd*) lpParameter)->WaitForData();
  return 0;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message) 
  {
  case WM_CREATE:
    mainwnd->Init(hWnd);
    return 0;
    break;
  case WM_SIZE:
    mainwnd->OnSize((DWORD) wParam, (UINT) (lParam&0xFFFF), (UINT) (lParam>>16));
    break;
  case WM_PAINT:
    {
      PAINTSTRUCT paint;
      BeginPaint(hWnd, &paint);
      mainwnd->OnDraw();
      EndPaint(hWnd, &paint);
    }
    break;
  case WM_TIMER:
    mainwnd->OnTimer(wParam);
    break;
  case WM_DESTROY:
    PostQuitMessage(0);
    break;
  case WM_CLOSE:
    DestroyWindow(hWnd);
    PostQuitMessage(0);
    break;
  case WM_LBUTTONDOWN:
    mainwnd->OnLButtonDown(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_MOUSEMOVE:
    mainwnd->OnMouseMove(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_LBUTTONUP:
    mainwnd->OnLButtonUp(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_MOUSEWHEEL:
    mainwnd->OnMouseWheel(wParam, lParam&0xFFFF, lParam>>16);
    return 0;
    break;
  case WM_KEYDOWN:
    mainwnd->OnKeyDown(wParam, lParam&0xFFFF, lParam>16);
    break;
  case WM_COMMAND:
    {
/*      WORD wmID=(WORD) (wParam&0xFFFF);
      WORD wmevent=(WORD) (wParam>>16);
      if (wmevent==CBN_SELCHANGE)
        mainwnd->OnComboChange();*/
    }
    break;
/*  case WM_KEYDOWN:
    mainwnd->OnKeyDown(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_KEYUP:
    mainwnd->OnKeyUp(wParam, lParam&0xFFFF, lParam>>16);
    break;*/
  case GLREDRAW:
    mainwnd->OnGLRedraw(lParam);
    break;
  default:
    return DefWindowProc(hWnd, message, wParam, lParam);
  }
  return DefWindowProc(hWnd, message, wParam, lParam);
}

void MainWnd::Init(HWND hwnd)
{
  this->hwnd=hwnd;

  setangle[0].SetText(L"Up");
  setangle[0].SetWindowPos(0, 10, 45, 30);

  setangle[1].SetText(L"Down");
  setangle[1].SetWindowPos(95, 10, 45, 30);

  swap.SetText(L"Swap Views");
  swap.SetWindowPos(150, 10, 90, 30);

  reset.SetText(L"Reset");
  reset.SetWindowPos(250, 10, 100, 30);

  track.SetText(L"Start Tracking");
  track.SetWindowPos(360, 10, 100, 30);

  record.SetText(L"Record");
  record.SetWindowPos(470, 10, 100, 30);

  playback.SetText(L"Playback");
  playback.SetWindowPos(580, 10, 100, 30);

  grab.SetText(L"Grab");
  grab.SetWindowPos(690, 10, 100, 30);

  HDC hdc=::GetDC(NULL);
  LOGFONT logfont;
  logfont.lfHeight=-(logfont.lfHeight=-MulDiv(10, GetDeviceCaps(hdc, LOGPIXELSY), 72));
  logfont.lfWidth=0;
  logfont.lfEscapement=0;
  logfont.lfOrientation=0;
  logfont.lfWeight=FW_NORMAL;
  logfont.lfItalic=FALSE;
  logfont.lfUnderline=FALSE;
  logfont.lfStrikeOut=FALSE;
  logfont.lfCharSet=DEFAULT_CHARSET;
  logfont.lfOutPrecision=OUT_DEFAULT_PRECIS;
  logfont.lfClipPrecision=CLIP_DEFAULT_PRECIS;
  logfont.lfQuality=DEFAULT_QUALITY;
  logfont.lfPitchAndFamily=DEFAULT_PITCH;
  wcscpy_s(logfont.lfFaceName, 32, L"Tahoma");
  HFONT hfont=CreateFontIndirect(&logfont);

  infos[0]=CreateWindowEx(WS_EX_CLIENTEDGE, L"EDIT", L"", WS_CHILD|WS_VISIBLE|WS_VSCROLL|ES_AUTOVSCROLL|ES_MULTILINE|ES_READONLY, 250, 350, 300, 100, hwnd, NULL, hinstance, NULL);
  SendMessage(infos[0], WM_SETFONT, (WPARAM) hfont, FALSE);
  SendMessage(infos[0], EM_SETLIMITTEXT, 0, 0);

  infos[1]=CreateWindowEx(WS_EX_CLIENTEDGE, L"EDIT", L"", WS_CHILD|WS_VISIBLE|WS_VSCROLL|ES_AUTOVSCROLL|ES_MULTILINE|ES_READONLY, 250, 350, 300, 100, hwnd, NULL, hinstance, NULL);
  SendMessage(infos[1], WM_SETFONT, (WPARAM) hfont, FALSE);
  SendMessage(infos[1], EM_SETLIMITTEXT, 0, 0);

  ::ReleaseDC(NULL, hdc);

  colprev.x=0;
  colprev.y=50;
  colprev.width=240;
  colprev.height=180;

  depthprev.x=0;
  depthprev.y=230;
  depthprev.width=240;
  depthprev.height=180;

  renderer[0].Init(hwnd, hinstance, 0, 410, 240, 180, true);
  renderer[1].Init(hwnd, hinstance, 250, 50, 300, 300, false);
  renderer[1].SetCameraTrack(&cameratrack);

  HDC hdcdst=::GetDC(hwnd);
  hdctmp=::CreateCompatibleDC(hdcdst);
  ::ReleaseDC(hwnd, hdcdst);

  InitKinect();
}

void MainWnd::OnSize(UINT nType, int cx, int cy)
{
  renderer[renderer[0].GetWindowPosX()<renderer[1].GetWindowPosX()?1:0].SetWindowPos(250, 50, cx-260, cy-190);
  SetWindowPos(infos[0], NULL, 250, cy-130, cx-500, 120, SWP_NOZORDER);
  SetWindowPos(infos[1], NULL, cx-250, cy-130, 240, 120, SWP_NOZORDER);
}

void MainWnd::OnTimer(UINT nIDEvent)
{

}

void MainWnd::SetNoOfThreads(int n)
{
  KillThreads();

  threadalive=true;
  noofthreads=n;
  jacobthreads=(HANDLE*) malloc(sizeof(HANDLE)*noofthreads);
  icperrorthreads=(HANDLE*) malloc(sizeof(HANDLE)*noofthreads);
  alerts=(HANDLE*) malloc(sizeof(HANDLE)*noofthreads);
  jacobthreadinfo=(JACOBIANTHREADINFO*) realloc(jacobthreadinfo, sizeof(JACOBIANTHREADINFO)*noofthreads);
  icpthreadinfo=(ICPERRORTHREADINFO*) realloc(icpthreadinfo, sizeof(ICPERRORTHREADINFO)*noofthreads);
  for (int i=0; i<noofthreads; i++)
  {
    jacobthreadinfo[i].alive=&threadalive;
    jacobthreadinfo[i].jacobian=jacobian;
    jacobthreadinfo[i].alerts[0]=CreateEvent(NULL, TRUE, FALSE, NULL);
    jacobthreadinfo[i].alerts[1]=CreateEvent(NULL, TRUE, FALSE, NULL);
    alerts[i]=jacobthreadinfo[i].alerts[1];
    jacobthreads[i]=CreateThread(NULL, 0, JacobianThread, jacobthreadinfo+i, 0, NULL);

    icpthreadinfo[i].alive=&threadalive;
    icpthreadinfo[i].alerts[0]=CreateEvent(NULL, TRUE, FALSE, NULL);
    icpthreadinfo[i].alerts[1]=alerts[i];
    icperrorthreads[i]=CreateThread(NULL, 0, ICPErrorThread, icpthreadinfo+i, 0, NULL);
  }

  pairmatch.SetNoOfThreads(n);
  pdupdate.SetNoOfThreads(n);
  clouddata.SetNoOfBuffers(n);
}

void MainWnd::KillThreads()
{
  if (noofthreads)
  {
    threadalive=false;
    for (int i=0; i<noofthreads; i++)
      SetEvent(jacobthreadinfo[i].alerts[0]);
  
    WaitForMultipleObjects(noofthreads, alerts, TRUE, INFINITE);

    for (int i=0; i<noofthreads; i++)
      ResetEvent(alerts[i]);

    for (int i=0; i<noofthreads; i++)
      SetEvent(icpthreadinfo[i].alerts[0]);

    WaitForMultipleObjects(noofthreads, alerts, TRUE, INFINITE);

    for (int i=0; i<noofthreads; i++)
    {
      CloseHandle(jacobthreadinfo[i].alerts[0]);
      CloseHandle(jacobthreadinfo[i].alerts[1]);

      CloseHandle(icpthreadinfo[i].alerts[0]);
    }

    free(jacobthreads);
    free(icperrorthreads);
    free(alerts);
    jacobthreads=NULL;
    icperrorthreads=NULL;
    alerts=NULL;
    noofthreads=0;
  }
}

void MainWnd::ClearInfoText(const int index)
{
  SetWindowText(infos[index], L"");
}

void MainWnd::AppendInfoText(const wchar_t* text, const int index)
{
  HWND info=infos[index];
  int textlen=GetWindowTextLength(info);
  SendMessage(info, EM_SETSEL, textlen, textlen);
  SendMessage(info, EM_REPLACESEL, FALSE, (LPARAM) text);

  int textlen2=GetWindowTextLength(info);
  if (textlen==textlen2)
  {
    SetWindowText(info, text);
    textlen2=wcslen(text);
  }

  SendMessage(info, EM_SETSEL, textlen2, textlen2);
}

void MainWnd::GetTime(wchar_t* buffer, const int bufflen)
{
  if (bufflen<11)
    return;
  else
  {
    time_t now;
    struct tm  tm_now;
    time(&now);
  
    localtime_s(&tm_now, &now);

    int pos=0;
    if (tm_now.tm_hour<10)
      buffer[pos++]='0';
    _itow_s(tm_now.tm_hour, buffer+pos, bufflen-pos, 10);

    buffer[2]=':';
    pos=3;
    if (tm_now.tm_min<10)
      buffer[pos++]='0';
    _itow_s(tm_now.tm_min, buffer+pos, bufflen-pos, 10);
    
    buffer[5]=':';
    pos=6;
    if (tm_now.tm_sec<10)
      buffer[pos++]='0';
    _itow_s(tm_now.tm_sec, buffer+pos, bufflen-pos, 10);
    buffer[8]=':';
    buffer[9]=' ';
    buffer[10]='\0';
  }
}

void MainWnd::WriteCameraTrackInfo(wchar_t* buffer, const int bufflen)
{
  int pos=3;
  wcscpy_s(buffer, bufflen, L"Rx=");
  pos+=FormatNumber(buffer+pos, bufflen-pos, cameratrack.rxyz[0]*180/M_PI, 3);

  wcscpy_s(buffer+pos, bufflen-pos, L"; Ry=");
  pos+=5;
  pos+=FormatNumber(buffer+pos, bufflen-pos, cameratrack.rxyz[1]*180/M_PI, 3);

  wcscpy_s(buffer+pos, bufflen-pos, L"; Rz=");
  pos+=5;
  pos+=FormatNumber(buffer+pos, bufflen-pos, cameratrack.rxyz[2]*180/M_PI, 3);

  wcscpy_s(buffer+pos, bufflen-pos, L"\r\nTx=");
  pos+=5;
  pos+=FormatNumber(buffer+pos, bufflen-pos, cameratrack.txyz[0], 3);

  wcscpy_s(buffer+pos, bufflen-pos, L"; Ty=");
  pos+=5;
  pos+=FormatNumber(buffer+pos, bufflen-pos, cameratrack.txyz[1], 3);

  wcscpy_s(buffer+pos, bufflen-pos, L"; Tz=");
  pos+=5;
  FormatNumber(buffer+pos, bufflen-pos, cameratrack.txyz[2], 3);
}

int MainWnd::FormatNumber(wchar_t* buffer, const int bufflen, double value, int dps)
{
  int pos=0;
  if (value<0)
  {
    buffer[0]='-';
    value=-value;
    pos=1;
  }

  int tval=(int) value;
  _itow_s(tval, buffer+pos, bufflen-pos, 10);
  pos+=(int) wcslen(buffer+pos);

  if (dps>0)
  {
    buffer[pos++]='.';
    while (dps>0)
    {
      value=(value-tval)*10;
      tval=(int) value;
      _itow_s(tval, buffer+pos, bufflen-pos, 10);
      pos+=(int) wcslen(buffer+pos);
      --dps;
    }
  }
  buffer[pos]='\0';
  return pos;
}

void MainWnd::OnLButtonDown(UINT nFlags, int x, int y)
{
  long oldval=sensorangle;

  if (x>colprev.x && x<colprev.x+colprev.width && y>colprev.y && y<colprev.y+colprev.height)
    SetCalibrationPoint(colprev, x, y, (nFlags&MK_CONTROL)==MK_CONTROL);
  else if (x>depthprev.x && x<depthprev.x+depthprev.width && y>depthprev.y && y<depthprev.y+depthprev.height)
    SetCalibrationPoint(depthprev, x, y, (nFlags&MK_CONTROL)==MK_CONTROL);
  else
  {
    if (swap.HitTest(x, y))
      SwapViews();
    else if (reset.HitTest(x, y))
      ResetGlobal();
    else if (track.HitTest(x, y))
      TrackData();
    else if (grab.HitTest(x, y))
      GrabData();
    else if (record.HitTest(x, y))
      Record();
    else if (playback.HitTest(x, y))
      Playback();
    else if (setangle[0].HitTest(x, y))
    {
      if (sensorangle<NUI_CAMERA_ELEVATION_MAXIMUM)
        ++sensorangle;
    }
    else if (setangle[1].HitTest(x, y))
    {
      if (sensorangle>NUI_CAMERA_ELEVATION_MINIMUM)
        --sensorangle;
    }

    if (sensorangle!=oldval)
    {
      long res=0;
      int attempts=0;
      do
      {
        sensor->NuiCameraElevationSetAngle(sensorangle);
        Sleep(200);
        sensor->NuiCameraElevationGetAngle(&res);
      } while (res!=sensorangle && attempts<10);
    
      sensorangle=res;

      HDC hdc=::GetDC(hwnd);
      DrawAngleControls(hdc, true);
      ::ReleaseDC(hwnd, hdc);
    }
  }
}

void MainWnd::OnMouseMove(UINT nFlags, int x, int y)
{
  if (nFlags&MK_LBUTTON)
  {
    if (x>colprev.x && x<colprev.x+colprev.width && y>colprev.y && y<colprev.y+colprev.height)
      SetCalibrationPoint(colprev, x, y, (nFlags&MK_CONTROL)==MK_CONTROL);
    else if (x>depthprev.x && x<depthprev.x+depthprev.width && y>depthprev.y && y<depthprev.y+depthprev.height)
      SetCalibrationPoint(depthprev, x, y, (nFlags&MK_CONTROL)==MK_CONTROL);
  }
}

void MainWnd::OnLButtonUp(UINT nFlags, int x, int y)
{
}

void MainWnd::OnMouseWheel(WPARAM wParam, int x, int y)
{
  RECT rect;
  ::GetWindowRect(hwnd, &rect);
  x-=rect.left+GetSystemMetrics(SM_CXSIZEFRAME);
  y-=rect.top+GetSystemMetrics(SM_CYSIZEFRAME)+GetSystemMetrics(SM_CYCAPTION);
  if (renderer[0].HitTest(x, y))
    renderer[0].OnMouseWheel(wParam, x, y);
  else if (renderer[1].HitTest(x, y))
    renderer[1].OnMouseWheel(wParam, x, y);
}

void MainWnd::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  if (nChar=='R')
  {
    renderer[0].ResetCamera(true);
    renderer[1].ResetCamera(true);
  }
  else if (nChar=='C')
  {
    for (int i=0; i<noofwldpts; i++)
    {
      worlddata[i].used=false;
      worlddata[i].matchpoint=NULL;
    }
  }
}

void MainWnd::OnDraw()
{
  HDC hdc=::GetDC(hwnd);
  swap.Draw(hdc);
  reset.Draw(hdc);
  track.Draw(hdc);
  record.Draw(hdc);
  playback.Draw(hdc);
  grab.Draw(hdc);
  DrawAngleControls(hdc, false);
  ::ReleaseDC(hwnd, hdc);

  OnGLRedraw(3);
}

void MainWnd::OnGLRedraw(LPARAM lParam)
{
  if (lParam&1)
    SendMessage(renderer[0].hwnd, GLREDRAW, 0, 0);
  if (lParam&2)
    SendMessage(renderer[1].hwnd, GLREDRAW, 0, 0);
}

void MainWnd::DrawAngleControls(HDC hdc, bool clear)
{
  if (clear)
  {
    RECT rect;
    rect.left=setangle[0].GetX();
    rect.right=setangle[1].GetX()+setangle[1].GetWidth();
    rect.top=setangle[0].GetY();
    rect.bottom=setangle[0].GetY()+setangle[0].GetHeight();

    ::FillRect(hdc, &rect, ::GetSysColorBrush(COLOR_3DFACE));
  }

  setangle[0].Draw(hdc);
  setangle[1].Draw(hdc);

  wchar_t buff[20];
  _itow_s(sensorangle, buff, 20, 10);
  wcscat_s(buff, 20, L" d");
  ::TextOut(hdc, 70, 15, buff, wcslen(buff));

  ::ReleaseDC(hwnd, hdc);
}

HRESULT MainWnd::InitKinect()
{
  HRESULT hr;

  sensorangle=0;
  colstream=depthstream=NULL;
  frameevents[0]=frameevents[1]=NULL;

  pointdata=(POINTDATA*) malloc(DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT*sizeof(POINTDATA));
  int pos=0;
  // starts at the bottom-left corner
  for (int j=0; j<DEPTH_FRAME_HEIGHT; j++)
  {
    for (int i=0; i<DEPTH_FRAME_WIDTH; i++)
    {
      pointdata[pos].xyz[0]=(i-(DEPTH_FRAME_WIDTH/2))*WORLD_SCALE;
      pointdata[pos].xyz[1]=(j-(DEPTH_FRAME_HEIGHT/2))*WORLD_SCALE;
      pointdata[pos++].used=false;
    }
  }

  HDC hdcsrc=::GetDC(NULL);
  BITMAPINFO bminfo;
  memset(&bminfo, 0, sizeof(BITMAPINFO));

  colprev.data=NULL;
  bminfo.bmiHeader.biSize=sizeof(BITMAPINFO);
  bminfo.bmiHeader.biWidth=COLOUR_FRAME_WIDTH;
  bminfo.bmiHeader.biHeight=COLOUR_FRAME_HEIGHT;
  bminfo.bmiHeader.biPlanes=1;
  bminfo.bmiHeader.biBitCount=32;
  bminfo.bmiHeader.biCompression=BI_RGB;
  bminfo.bmiHeader.biSizeImage=bminfo.bmiHeader.biWidth*bminfo.bmiHeader.biHeight*bminfo.bmiHeader.biBitCount/8;
  colprev.hbmpprev=::CreateDIBSection(hdcsrc, &bminfo, DIB_RGB_COLORS, (void**) &colprev.data, NULL, NULL);
  colprev.srcwidth=COLOUR_FRAME_WIDTH;
  colprev.srcheight=COLOUR_FRAME_HEIGHT;
  
  depthprev.data=NULL;
  bminfo.bmiHeader.biWidth=DEPTH_FRAME_WIDTH;
  bminfo.bmiHeader.biHeight=DEPTH_FRAME_HEIGHT;
  bminfo.bmiHeader.biPlanes=1;
  bminfo.bmiHeader.biBitCount=32;
  bminfo.bmiHeader.biCompression=BI_RGB;
  bminfo.bmiHeader.biSizeImage=bminfo.bmiHeader.biWidth*bminfo.bmiHeader.biHeight*bminfo.bmiHeader.biBitCount/8;
  depthprev.hbmpprev=::CreateDIBSection(hdcsrc, &bminfo, DIB_RGB_COLORS, (void**) &depthprev.data, NULL, NULL);
  depthprev.srcwidth=DEPTH_FRAME_WIDTH;
  depthprev.srcheight=DEPTH_FRAME_HEIGHT;

  ::ReleaseDC(NULL, hdcsrc);

#ifdef NEAR_MODE
  // Near mode
  DWORD flags=NUI_IMAGE_STREAM_FLAG_ENABLE_NEAR_MODE|NUI_IMAGE_STREAM_FLAG_DISTINCT_OVERFLOW_DEPTH_VALUES;
  depthrange[0]=NUI_IMAGE_DEPTH_MINIMUM_NEAR_MODE;
  depthrange[1]=NUI_IMAGE_DEPTH_MAXIMUM_NEAR_MODE;
#else
  // Default mode
  DWORD flags=NUI_IMAGE_STREAM_FLAG_DISTINCT_OVERFLOW_DEPTH_VALUES;
  depthrange[0]=NUI_IMAGE_DEPTH_MINIMUM;
  depthrange[1]=NUI_IMAGE_DEPTH_MAXIMUM;
#endif

  depthrange[0]>>=NUI_IMAGE_PLAYER_INDEX_SHIFT;
  depthrange[1]>>=NUI_IMAGE_PLAYER_INDEX_SHIFT;
  depthrange[2]=depthrange[1]-depthrange[0];

  renderer[0].SetCameraNearFar(CalcDataDepth(depthrange[0]), CalcDataDepth(depthrange[1]), DEPTH_FRAME_WIDTH, DEPTH_FRAME_HEIGHT);
  renderer[1].SetCameraNearFar(CalcDataDepth(depthrange[0]), CalcDataDepth(depthrange[1]), DEPTH_FRAME_WIDTH, DEPTH_FRAME_HEIGHT);

  int iSensorCount=0;
  hr=NuiGetSensorCount(&iSensorCount);
  if (FAILED(hr))
    return hr;

  for (int i=0; i<iSensorCount; i++)
  {
    sensor=NULL;
    if (NuiCreateSensorByIndex(i, &sensor)==S_OK && sensor->NuiStatus()==S_OK)
    {
      if (sensor->NuiInitialize(NUI_INITIALIZE_FLAG_USES_COLOR|NUI_INITIALIZE_FLAG_USES_DEPTH)==S_OK)
      {
//        sensorangle=11;
//        sensor->NuiCameraElevationSetAngle(sensorangle);

        kinectalive=true;
        CreateThread(NULL, 0, ListenThread, this, 0, NULL);

        frameevents[0]=CreateEvent(NULL, TRUE, FALSE, NULL);
        frameevents[1]=CreateEvent(NULL, TRUE, FALSE, NULL);

        hr=sensor->NuiImageStreamOpen(NUI_IMAGE_TYPE_COLOR, IMAGE_RESOLUTION, 0, 2, frameevents[0], &colstream);
        hr=sensor->NuiImageStreamOpen(NUI_IMAGE_TYPE_DEPTH, DEPTH_RESOLUTION, flags, 2, frameevents[1], &depthstream);

        sensor->NuiCameraElevationGetAngle(&sensorangle);
        return hr;
      }
    }
    if (sensor)
    {
      sensor->Release();
      sensor=NULL;
    }
  }
  return hr;
}

void MainWnd::ProcessColour()
{
  if (playing)
  {
    ResetEvent(frameevents[0]);
    return;
  }

  // Kinect data starts from the top-right corner
  // hbmp starts from the lower-left corner

  NUI_IMAGE_FRAME imageFrame;
  HRESULT hr=sensor->NuiImageStreamGetNextFrame(colstream, 0, &imageFrame);
  if (hr==S_OK)
  {
    NUI_LOCKED_RECT frameinfo;
    imageFrame.pFrameTexture->LockRect(0, &frameinfo, NULL, 0);

    if (frameinfo.Pitch)
    {
      frameinfo.size=frameinfo.size;
      int width=colprev.srcwidth;
      int height=colprev.srcheight;
      DWORD* src=((DWORD*) frameinfo.pBits)+width*height-1;
      DWORD* dst=colprev.data;
      for (int i=width*height; i>0; i--)
        *(dst++)=*(src--);
/*      for (int j=0; j<height; j++)
      {
        for (int i=0; i<width; i++)
        {
          DWORD col=*src;
          *dst=col;

          ++dst;
          --src;
        }
      }*/

      HDC hdcdst=::GetDC(hwnd);
      ::SelectObject(hdctmp, colprev.hbmpprev);
      ::SetStretchBltMode(hdcdst, COLORONCOLOR);//HALFTONE);
//      ::SetBrushOrgEx(hdcdst, 0, 0, NULL);
      ::StretchBlt(hdcdst, colprev.x, colprev.y, colprev.width, colprev.height, hdctmp, 0, 0, colprev.srcwidth, colprev.srcheight, SRCCOPY);

      UpdateFPS(colprev);
      ::SetTextAlign(hdcdst, TA_TOP|TA_RIGHT);
      ::SetTextColor(hdcdst, 0x00FFFFFF);
      ::SetBkMode(hdcdst, TRANSPARENT);
      ::TextOut(hdcdst, colprev.x+colprev.width, colprev.y, colprev.lastfps, wcslen(colprev.lastfps));

      DrawCalibrationPoints(hdcdst, depthprev);

      ::ReleaseDC(hwnd, hdcdst);
    }

    imageFrame.pFrameTexture->UnlockRect(0);
    sensor->NuiImageStreamReleaseFrame(colstream, &imageFrame);
  }

  ResetEvent(frameevents[0]);
}

void MainWnd::DrawCalibrationPoints(HDC hdc, PREVIEWINFO& preview)
{
/*  float scalex=preview.width/(float) preview.srcwidth;
  float scaley=preview.height/(float) preview.srcheight;

  RECT rect;
  for (int i=0; i<2; i++)
  {
    rect.left=preview.x+(int) (preview.calibration[i].x*scalex)-2;
    rect.top=preview.y+(int) (preview.calibration[i].y*scaley)-2;
    rect.right=rect.left+5;
    rect.bottom=rect.top+5;
    ::FillRect(hdc, &rect, (HBRUSH) ::GetStockObject(BLACK_BRUSH));
    ++rect.left;
    ++rect.top;
    --rect.right;
    --rect.bottom;
    ::FillRect(hdc, &rect, (HBRUSH) ::GetStockObject(WHITE_BRUSH));
  }
  */
}

void MainWnd::UpdateFPS(PREVIEWINFO& info)
{
  ++info.fps;
  long now=::GetTickCount();
  long elapsed=now-info.counter;
  if (elapsed>1000)
  {
    int tmp=((1000*info.fps)/elapsed);
    _itow_s(tmp, info.lastfps, 20, 10);
    info.counter=now;
    info.fps=0;
  }
}

void MainWnd::ProcessDepth()
{
  if (playing)
  {
    ResetEvent(frameevents[1]);
    return;
  }

  NUI_IMAGE_FRAME imageFrame;
  HRESULT hr=sensor->NuiImageStreamGetNextFrame(depthstream, 0, &imageFrame);
  if (hr==S_OK)
  {
    NUI_LOCKED_RECT frameinfo;
    imageFrame.pFrameTexture->LockRect(0, &frameinfo, NULL, 0);

    if (frameinfo.Pitch)
    {
      frameinfo.size=frameinfo.size;
      int width=depthprev.srcwidth;
      int height=depthprev.srcheight;
      WORD* src=((WORD*) frameinfo.pBits)+width*height-1;
      DWORD* dst=depthprev.data;
      POINTDATA* nextpoint=pointdata;
      int used=0;

      for (int j=0; j<height; j++)
      {
        for (int i=0; i<width; i++)
        {
          nextpoint->used=false;
          if (*src==NUI_DEPTH_DEPTH_UNKNOWN_VALUE)
            *dst=UNKNOWN_COL;
          else if (*src==NUI_IMAGE_DEPTH_NO_VALUE)
            *dst=TOO_CLOSE_COL;
          else if (*src==NUI_IMAGE_DEPTH_TOO_FAR_VALUE)
            *dst=TOO_FAR_COL;
          else
          {
            //int ni=(int) (calibration[0]*i+calibration[3]*j+calibration[6]);
            //int nj=(int) (calibration[1]*i+calibration[4]*j+calibration[7]);
          
            long ni, nj;
            if (sensor->NuiImageGetColorPixelCoordinatesFromDepthPixelAtResolution(IMAGE_RESOLUTION, DEPTH_RESOLUTION, NULL, i, j, *src, &ni, &nj)!=S_OK)
              ni=nj=-1;

//            if (j<10 || j>height-10 || i<10 || i>width-10)
//              ni=nj=-1;

            if (ni>=0 && ni<colprev.srcwidth && nj>=0 && nj<colprev.srcheight)
            {
              WORD depth=(*src)>>NUI_IMAGE_PLAYER_INDEX_SHIFT;

              nextpoint->paired=false;
              nextpoint->nxyz[0]=nextpoint->nxyz[1]=nextpoint->nxyz[2]=0.01f;

              float scale=depth*PERSPECTIVE_CORRECTION*WORLD_SCALE;
              nextpoint->xyz[0]=(i-(DEPTH_FRAME_WIDTH/2))*scale;//WORLD_SCALE;
              nextpoint->xyz[1]=(j-(DEPTH_FRAME_HEIGHT/2))*scale;//WORLD_SCALE;
              nextpoint->xyz[2]=CalcDataDepth(depth);//Z_OFFSET-depth*WORLD_SCALE*DEPTH_SCALE;

              DWORD col=colprev.data[nj*colprev.srcwidth+ni];
              nextpoint->colour[0]=((col>>16)&0xFF)/256.0f;
              nextpoint->colour[1]=((col>>8)&0xFF)/256.0f;
              nextpoint->colour[2]=(col&0xFF)/256.0f;

              ++used;
              nextpoint->used=true;

              if (recorder)
                recorder->WritePoint(nextpoint, nextpoint-pointdata);

              BYTE val=(BYTE) ((256*(depth-depthrange[0]))/depthrange[2]);
              *dst=val|(val<<8)|(val<<16);

              assert(depth>=depthrange[0] && depth<=depthrange[1]);
            }
            else
              *dst=UNKNOWN_COL;
          }

          --src;
          ++dst;
          ++nextpoint;
        }
      }

      if (recorder && !recording)
      {
        recorder->CloseRecorder();
        delete recorder;
        recorder=NULL;
      }

      if (recorder && !worlddata)
        TrackData(pointdata, used);

      if (!recorder && tracking)// || !worlddata)
      {
/*        if (GetFileAttributes(L"TrackDump.dat")==INVALID_FILE_ATTRIBUTES)
        {
          FILE* file=NULL;
          _wfopen_s(&file, L"TrackDump.dat", L"wb");
          fwrite(&used, sizeof(int), 1, file);
          fwrite(pointdata, sizeof(POINTDATA), FRAME_WIDTH*FRAME_HEIGHT, file);
          fclose(file);
        }
        else
        {
          FILE* file=NULL;
          _wfopen_s(&file, L"TrackDump.dat", L"rb");
          fread(&used, sizeof(int), 1, file);
          fread(pointdata, sizeof(POINTDATA), FRAME_WIDTH*FRAME_HEIGHT, file);
          fclose(file);
        }*/

        TrackData(pointdata, used);
      }

      HDC hdcdst=::GetDC(hwnd);
      ::SelectObject(hdctmp, depthprev.hbmpprev);
      ::SetStretchBltMode(hdcdst, COLORONCOLOR);//HALFTONE);
      //::SetBrushOrgEx(hdcdst, 0, 0, NULL);
      ::StretchBlt(hdcdst, depthprev.x, depthprev.y, depthprev.width, depthprev.height, hdctmp, 0, 0, depthprev.srcwidth, depthprev.srcheight, SRCCOPY);

      UpdateFPS(depthprev);
      ::SetTextAlign(hdcdst, TA_TOP|TA_RIGHT);
      ::SetTextColor(hdcdst, 0x00808000);
      ::SetBkMode(hdcdst, TRANSPARENT);
      ::TextOut(hdcdst, depthprev.x+depthprev.width, depthprev.y, depthprev.lastfps, wcslen(depthprev.lastfps));

      DrawCalibrationPoints(hdcdst, colprev);

      ::ReleaseDC(hwnd, hdcdst);
    }

    imageFrame.pFrameTexture->UnlockRect(0);
    sensor->NuiImageStreamReleaseFrame(depthstream, &imageFrame);
  }

  if (!recording)
    renderer[0].SetPointData(pointdata, DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT, true);

  if (exportdata[0])
  {
    CreatePolygonModel(true);
    ExportWRLModel(true);
  }
/*  else if (exportdata[1])
  {
    CreatePolygonModel(false);
    ExportWRLModel(false);
  }*/

  ResetEvent(frameevents[1]);
}

void MainWnd::WaitForData()
{
  while (kinectalive)
  {
    DWORD res=MsgWaitForMultipleObjects(2, frameevents, false, INFINITE, 0);
    if (kinectalive)
    {
      DWORD index=res-WAIT_OBJECT_0;
      if (index==1)
        ProcessDepth();
      else if (index==0)
        ProcessColour();
    }
  }
}

void MainWnd::SetCalibrationPoint(PREVIEWINFO& preview, int x, int y, bool second)
{
  int index=(second?1:0);
  float scalex=preview.srcwidth/(float) preview.width;
  float scaley=preview.srcheight/(float) preview.height;
  preview.calibration[index].x=(LONG) ((x-preview.x)*scalex);
  preview.calibration[index].y=(LONG) ((y-preview.y)*scaley);

  UpdateCalibration();
}

void MainWnd::UpdateCalibration()
{
  bool calibrate=true;
  for (int i=0; i<2 && calibrate; i++)
    if (!colprev.calibration[i].x || !colprev.calibration[i].y || !depthprev.calibration[i].x || !depthprev.calibration[i].y)
      calibrate=false;

  if (calibrate)
  {
    calibration[0]=(colprev.calibration[0].x-colprev.calibration[1].x)/(float) (depthprev.calibration[0].x-depthprev.calibration[1].x);
    calibration[6]=colprev.calibration[0].x-depthprev.calibration[0].x*calibration[0];
    
    calibration[4]=(colprev.calibration[0].y-colprev.calibration[1].y)/(float) (depthprev.calibration[0].y-depthprev.calibration[1].y);
    calibration[7]=colprev.calibration[0].y-depthprev.calibration[0].y*calibration[4];
  }
}

void MainWnd::SaveSettings()
{
  FILE* file=NULL;
  if (!_wfopen_s(&file, SETTINGS_FILE, L"wb") && file)
  {
    fwrite(colprev.calibration, sizeof(POINT), 2, file);
    fwrite(depthprev.calibration, sizeof(POINT), 2, file);
    fclose(file);
  }
}

void MainWnd::LoadSettings()
{
  FILE* file=NULL;
  if (!_wfopen_s(&file, SETTINGS_FILE, L"rb") && file)
  {
    fread(colprev.calibration, sizeof(POINT), 2, file);
    fread(depthprev.calibration, sizeof(POINT), 2, file);
    fclose(file);
  }
  UpdateCalibration();
}

void MainWnd::SwapViews()
{
  int x1, y1, w1, h1, x2, y2, w2, h2;
  renderer[0].GetWindowPos(x1, y1, w1, h1);
  renderer[1].GetWindowPos(x2, y2, w2, h2);

  renderer[0].SetWindowPos(x2, y2, w2, h2);
  renderer[1].SetWindowPos(x1, y1, w1, h1);
}

void MainWnd::ResetGlobal()
{
  if (MessageBox(hwnd, L"Reset Global View?", L"Kinecting Up", MB_ICONWARNING|MB_YESNO)==IDYES)
  {
    renderer[0].SetPointData(NULL, 0, true);
    renderer[1].SetPointData(NULL, 0, true);
    clouddata.ClearPointData();
    free(worlddata);
    worlddata=NULL;
    maxwldpts=noofwldpts=0;

    trackedframes=0;
    lastgrab=0;
    lastpruned=0;

    pairratio[0]=100.0f;
    pairratio[1]=100.0f;
  }
}
/*
void MainWnd::DumpDepthAnalysis()
{
  char cbuffer[MAX_PATH];
  wchar_t wbuffer[MAX_PATH];
  FILE* file=NULL;
  if (!_wfopen_s(&file, MIN_MAX_FILE, L"wb") && file)
  {
    int pos=0;
    float max=0;
    for (int j=0; j<DEPTH_FRAME_HEIGHT; j++)
    {
      for (int i=0; i<DEPTH_FRAME_WIDTH; i++, pos++)
      {
        float val=(depthseen[pos]?depthmaxs[pos]-depthmins[pos]:0);
        if (val>max)
          max=val;
        FormatNumber(wbuffer, MAX_PATH, val, 5);
        WideCharToMultiByte(CP_ACP, 0, wbuffer, -1, cbuffer, MAX_PATH, 0, NULL);
        fwrite(cbuffer, 1, strlen(cbuffer), file);
        fwrite(", ", 1, 2, file);
      }
      fwrite("\r\n", 1, 2, file);
    }
    FormatNumber(wbuffer, MAX_PATH, max, 5);
    WideCharToMultiByte(CP_ACP, 0, wbuffer, -1, cbuffer, MAX_PATH, 0, NULL);
    fwrite(cbuffer, 1, strlen(cbuffer), file);
    fclose(file);
  }

  file=NULL;
  if (!_wfopen_s(&file, AVERAGE_FILE, L"wb") && file)
  {
    int pos=0;
    for (int j=0; j<DEPTH_FRAME_HEIGHT; j++)
    {
      for (int i=0; i<DEPTH_FRAME_WIDTH; i++, pos++)
      {
        FormatNumber(wbuffer, MAX_PATH, -depthavgs[pos], 5);
        WideCharToMultiByte(CP_ACP, 0, wbuffer, -1, cbuffer, MAX_PATH, 0, NULL);
        fwrite(cbuffer, 1, strlen(cbuffer), file);
        fwrite(", ", 1, 2, file);
      }
      fwrite("\r\n", 1, 2, file);
    }
    fclose(file);
  }

  int counts[4000];
  float differences[4000];

  for (int i=0; i<4000; i++)
  {
    counts[i]=0;
    differences[i]=0;
  }

  for (int i=0; i<DEPTH_FRAME_HEIGHT*DEPTH_FRAME_HEIGHT; i++)
  {
    if (depthseen[i])
    {
      int index=(int) (-depthavgs[i]*1000);
      assert(index>=0 && index<4000);

      ++counts[index];
      differences[index]+=(depthmaxs[i]-depthmins[i])*1000.0f;

//      int percentage=(100*depthcounts[i])/depthframes;
//      counts[percentage]++;
    }
  }

  file=NULL;
  if (!_wfopen_s(&file, DEPTH_FILE, L"wb") && file)
  {
    for (int i=0; i<=100; i++)
    {
//      if (counts[i])
      {
//        float val=100*differences[i]/counts[i];

//        if (val<500)
        {
          _itoa_s(i, cbuffer, MAX_PATH, 10);
          fwrite(cbuffer, 1, strlen(cbuffer), file);
          fwrite(", ", 1, 2, file);

//          FormatNumber(wbuffer, MAX_PATH, val, 5);
//          WideCharToMultiByte(CP_ACP, 0, wbuffer, -1, cbuffer, MAX_PATH, 0, NULL);
          _itoa_s(counts[i], cbuffer, MAX_PATH, 10);
          fwrite(cbuffer, 1, strlen(cbuffer), file);
          fwrite("\r\n", 1, 2, file);
        }
      }
    }
    fclose(file);
  }

}*/

void MainWnd::TrackData()
{
  if (tracking)
  {
    track.SetText(L"Start Tracking");
    tracking=false;

//    DumpDepthAnalysis();
  }
  else
  {
//    ClearInfoText(0);
    track.SetText(L"Stop Tracking");
    tracking=true;
  }

  HDC hdc=::GetDC(hwnd);
  RECT rect;
  rect.left=track.GetX();
  rect.right=rect.left+track.GetWidth();
  rect.top=track.GetY();
  rect.bottom=rect.top+track.GetHeight();

  ::FillRect(hdc, &rect, ::GetSysColorBrush(COLOR_3DFACE));
  track.Draw(hdc);

  ::ReleaseDC(hwnd, hdc);
}

void MainWnd::Record()
{
  if (recorder)
  {
    record.SetText(L"Record");
    recording=false;
  }
  else
  {
    record.SetText(L"Stop");
    recording=true;
    recorder=new Recorder(RECORD_FILE, RECORDER_RECORD, DEPTH_FRAME_WIDTH, DEPTH_FRAME_HEIGHT);
  }

  HDC hdc=::GetDC(hwnd);
  RECT rect;
  rect.left=record.GetX();
  rect.right=rect.left+record.GetWidth();
  rect.top=track.GetY();
  rect.bottom=rect.top+record.GetHeight();

  ::FillRect(hdc, &rect, ::GetSysColorBrush(COLOR_3DFACE));
  record.Draw(hdc);

  ::ReleaseDC(hwnd, hdc);
}

void MainWnd::Playback()
{
  if (recorder)
  {
    playing=false;
    playback.SetText(L"Playback");
//    DumpDepthAnalysis();
//    SetWindowText(hwnd, L"Kinecting up the Past");
  }
  else
  {
    playing=true;
    playback.SetText(L"Stop");
    recorder=new Recorder(RECORD_FILE, RECORDER_PLAYBACK, DEPTH_FRAME_WIDTH, DEPTH_FRAME_HEIGHT);
    CreateThread(NULL, 0, PlaybackThread, this, 0, NULL);
  }

  HDC hdc=::GetDC(hwnd);
  RECT rect;
  rect.left=playback.GetX();
  rect.right=rect.left+playback.GetWidth();
  rect.top=playback.GetY();
  rect.bottom=rect.top+playback.GetHeight();

  ::FillRect(hdc, &rect, ::GetSysColorBrush(COLOR_3DFACE));
  playback.Draw(hdc);

  ::ReleaseDC(hwnd, hdc);
}

void MainWnd::GrabData()
{
//  grabdata=true;
  if (MessageBox(hwnd, L"View Point Pairs?", L"Kinecting up the Past", MB_ICONQUESTION|MB_YESNO)==IDYES)
    ShellExecute(NULL, L"open", L"PairViewer.exe", MATCH_FILE, NULL , SW_SHOW);
  else if (MessageBox(hwnd, L"Export Frame Data?", L"Kinecting up the Past", MB_ICONQUESTION|MB_YESNO)==IDYES)
    exportdata[0]=true;
  else if (MessageBox(hwnd, L"Export World Data?", L"Kinecting up the Past", MB_ICONQUESTION|MB_YESNO)==IDYES)
  {
//    exportdata[1]=true;
    CreatePolygonModel(false);
    ExportWRLModel(false);
  }
  else
  {
    MessageBox(hwnd, L"Grabbing screenshot", L"Kinecting up the Past", MB_ICONINFORMATION);
    Sleep(200);
    ScreenGrab();
  }
}

void MainWnd::InitTrackingMemory()
{
  int maxsize=DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT;
  r=(double*) malloc(sizeof(double)*maxsize);
  jacobian=(double*) malloc(sizeof(double)*maxsize*6);
  tmpsqr=(double*) malloc(sizeof(double)*6*6);
  lowermat=(double*) malloc(sizeof(double)*6*6);
  b=(double*) malloc(sizeof(double)*6);
  y=(double*) malloc(sizeof(double)*6);
  pairs=(PAIR*) malloc(sizeof(PAIR)*maxsize);
}

void MainWnd::FreeTrackingMemory()
{
  free(r);
  free(jacobian);
  free(tmpsqr);
  free(lowermat);
  free(b);
  free(y);
  free(pairs);

  r=jacobian=tmpsqr=lowermat=b=y=NULL;
  pairs=NULL;
}

void MainWnd::CalcNormals(POINTDATA* frame, int& used)
{
  int pos=0;

  int noofrows=DEPTH_FRAME_HEIGHT-1;
  int noofcols=DEPTH_FRAME_WIDTH-1;

  POINTDATA* line1=pointdata;
  POINTDATA* line2=pointdata+DEPTH_FRAME_WIDTH;

//  const float nz=WORLD_SCALE*WORLD_SCALE;
//  const float nz2=nz*nz;
//  const float threshold=2*nz2;

  for (int j=0; j<noofrows; j++)
  {
    for (int i=0; i<noofcols; i++)
    {
      if (line1[i].used)
      {
        if (line1[i+1].used && line2[i].used)
        {
          float ax=line1[i+1].xyz[0]-line1[i].xyz[0];
          float bx=line2[i].xyz[0]-line1[i].xyz[0];

          float ay=line1[i+1].xyz[1]-line1[i].xyz[1];
          float by=line2[i].xyz[1]-line1[i].xyz[1];

          float az=line1[i+1].xyz[2]-line1[i].xyz[2];
          float bz=line2[i].xyz[2]-line1[i].xyz[2];

//          if (d1>-WORLD_SCALE && d1<WORLD_SCALE && d2>-WORLD_SCALE && d2<WORLD_SCALE)
          {
            double nx=ay*bz-az*by;
            double ny=az*bx-ax*bz;
            double nz=ax*by-ay*bx;
//            float nx=-d1*WORLD_SCALE;
//            float ny=-d2*WORLD_SCALE;
            double mag=nx*nx+ny*ny+nz*nz;
/*            if (mag<1e9)
            {
              --used;
              pointdata[pos].used=false;
            }
            else*/
            {
              mag=sqrt(mag);

              pointdata[pos].nxyz[0]=(float) (nx/mag);
              pointdata[pos].nxyz[1]=(float) (ny/mag);
              pointdata[pos].nxyz[2]=(float) (nz/mag);

//            pointdata[pos].matchpoint=NULL;
//            pointdata[pos].matchdist=DISTANCE_THRESHOLD*DISTANCE_THRESHOLD*100;
            }
          }
/*          else
          {
            --used;
            pointdata[pos].used=false;
          }*/
        }
        else
        {
          --used;
          pointdata[pos].used=false;
        }
      }
      ++pos;
    }
    if (pointdata[pos].used)
    {
      --used;
      pointdata[pos].used=false;
    }
    ++pos;

    line1+=DEPTH_FRAME_WIDTH;
    line2+=DEPTH_FRAME_WIDTH;
  }

  for (int i=0; i<DEPTH_FRAME_WIDTH; i++)
  {
    if (pointdata[pos].used)
    {
      --used;
      pointdata[pos].used=false;
    }
    ++pos;
  }

}

void MainWnd::TrackData(POINTDATA* frame, int used)
{
#ifdef CLEAR_WORLD_PAIRS
  for (int i=0; i<noofwldpts; i++)
    worlddata[i].paired=false;
#endif

  ++trackedframes;

  CalcNormals(frame, used);

  if (!worlddata)
  {
/*    depthframes=0;
    for (int i=0; i<DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT; i++)
    {
      depthseen[i]=false;
      depthcounts[i]=0;
    }*/

    noofwldpts=0;
    maxwldpts=10*DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT;
    worlddata=(POINTDATA*) malloc(sizeof(POINTDATA)*maxwldpts);

    memset(&cameratrack, 0, sizeof(CAMERATRACK));
    cameratrack.txyz[0]=cameratrack.txyz[1]=cameratrack.txyz[2]=0.001f;
    cameratrack.rxyz[0]=cameratrack.rxyz[1]=cameratrack.rxyz[2]=(float) (M_PI/180)*0.002f;

    CalcCosSins(cameratrack.coss, cameratrack.sins, cameratrack.rxyz);

    double transform[16];
    CalcTransform(transform, cameratrack.coss, cameratrack.sins, cameratrack.txyz);
    for (int i=0; i<DEPTH_FRAME_WIDTH*(DEPTH_FRAME_HEIGHT-1); i++)
    {
      if (pointdata[i].used)
      {
        POINTDATA* pt=pointdata+i;
        worlddata[noofwldpts]=pointdata[i];
        
        worlddata[noofwldpts].xyz[0]=(float) (transform[0]*pt->xyz[0]+transform[4]*pt->xyz[1]+transform[8]*pt->xyz[2]+transform[12]);
        worlddata[noofwldpts].xyz[1]=(float) (transform[1]*pt->xyz[0]+transform[5]*pt->xyz[1]+transform[9]*pt->xyz[2]+transform[13]);
        worlddata[noofwldpts].xyz[2]=(float) (transform[2]*pt->xyz[0]+transform[6]*pt->xyz[1]+transform[10]*pt->xyz[2]+transform[14]);

        worlddata[noofwldpts].nxyz[0]=(float) (transform[0]*pt->nxyz[0]+transform[4]*pt->nxyz[1]+transform[8]*pt->nxyz[2]);
        worlddata[noofwldpts].nxyz[1]=(float) (transform[1]*pt->nxyz[0]+transform[5]*pt->nxyz[1]+transform[9]*pt->nxyz[2]);
        worlddata[noofwldpts].nxyz[2]=(float) (transform[2]*pt->nxyz[0]+transform[6]*pt->nxyz[1]+transform[10]*pt->nxyz[2]);

        worlddata[noofwldpts].matchpoint=NULL;
        worlddata[noofwldpts].lastseen=trackedframes;
        clouddata.AddPointData(worlddata+noofwldpts);
        worlddata[noofwldpts++].count=1;
      }
    }
    DeleteFile(MATCH_FILE);

    assert(clouddata.CheckConsistency());

    renderer[1].SetPointData(worlddata, noofwldpts, true);

//    TrackDataAux(frame, used);
//    TrackData();
//    Playback();
  }
  else
  {
/*    ++depthframes;
    for (int i=0; i<DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT; i++)
    {
      if (pointdata[i].used)
      {
        if (depthseen[i])
        {
          if (pointdata[i].xyz[2]<depthmins[i])
            depthmins[i]=pointdata[i].xyz[2];
          else if (pointdata[i].xyz[2]>depthmaxs[i])
            depthmaxs[i]=pointdata[i].xyz[2];

          float avg=(pointdata[i].xyz[2]+depthavgs[i]*depthcounts[i])/(depthcounts[i]+1);
          depthavgs[i]=avg;
          ++depthcounts[i];
        }
        else
        {
          depthseen[i]=true;
          depthavgs[i]=depthmaxs[i]=depthmins[i]=pointdata[i].xyz[2];
          depthcounts[i]=1;
        }
      }
    }*/

    TrackDataAux(frame, used);
  }

//  ScreenGrab();
}

void MainWnd::CalcFrustum(float* f_near, float* f_far)
{
  float scale=depthrange[0]*PERSPECTIVE_CORRECTION*WORLD_SCALE;
  f_near[0]=(DEPTH_FRAME_WIDTH/2)*scale;
  f_near[1]=(DEPTH_FRAME_HEIGHT/2)*scale;
  f_near[2]=CalcDataDepth(depthrange[0]);

  scale=depthrange[1]*PERSPECTIVE_CORRECTION*WORLD_SCALE;
  f_far[0]=(DEPTH_FRAME_WIDTH/2)*scale;
  f_far[1]=(DEPTH_FRAME_HEIGHT/2)*scale;
  f_far[2]=CalcDataDepth(depthrange[1]);
}

void MainWnd::CalcFrameMinMax(const double* transform, float* min, float* max)
{
  float scale=depthrange[1]*PERSPECTIVE_CORRECTION*WORLD_SCALE;
  min[0]=-(DEPTH_FRAME_WIDTH/2)*scale;
  min[1]=-(DEPTH_FRAME_HEIGHT/2)*scale;
  min[2]=CalcDataDepth(depthrange[1]);

  max[0]=-min[0];
  max[1]=-min[1];
  max[2]=CalcDataDepth(depthrange[0]);

  int i;
  float pts_in[4*8], pts_out[4*8];
  pts_in[0]=min[0];
  pts_in[1]=min[1];
  pts_in[2]=min[2];
  pts_in[3]=1.0f;

  pts_in[4]=max[0];
  pts_in[5]=min[1];
  pts_in[6]=min[2];
  pts_in[7]=1.0f;

  pts_in[8]=max[0];
  pts_in[9]=max[1];
  pts_in[10]=min[2];
  pts_in[11]=1.0f;

  pts_in[12]=min[0];
  pts_in[13]=max[1];
  pts_in[14]=min[2];
  pts_in[15]=1.0f;

  for (i=0; i<16; i++)
    pts_in[i+16]=pts_in[i];

  pts_in[4*4+2]=max[2];
  pts_in[5*4+2]=max[2];
  pts_in[6*4+2]=max[2];
  pts_in[7*4+2]=max[2];

  MatMult(pts_out, transform, 4, 4, pts_in, 4, 8);

  for (i=0; i<3; i++)
    min[i]=max[i]=pts_out[i];
  min[3]=max[3]=1.0f;

  for (int i=1; i<8; i++)
  {
    int index=4*i;
    for (int j=0; j<3; j++, ++index)
    {
      if (pts_out[index]<min[j])
        min[j]=pts_out[index];
      if (pts_out[index]>max[j])
        max[j]=pts_out[index];
    }
  }
}

void MainWnd::TrackDataInner(POINTDATA* frame, const int used, const int noofpairs)
{
  DWORD start=::GetTickCount();
  int i, j, iteration=0;
  double coss[3], sins[3], variables[6], transform[16], update[6];

  InitICPErrorThreadMemory(transform, noofpairs, pairs);
  InitJacobianThreadMemory(coss, sins, noofpairs, pairs);

  for (i=0; i<3; i++)
  {
    variables[i]=cameratrack.rxyz[i];
    variables[3+i]=cameratrack.txyz[i];

    coss[i]=cameratrack.coss[i];
    sins[i]=cameratrack.sins[i];
  }

  CalcTransform(transform, coss, sins, cameratrack.txyz);

  double besterror=CalcInnerICPError();

  do
  {
    CalcJacobian();

    MatMultTA_LT(tmpsqr, jacobian, noofpairs, 6, jacobian, noofpairs, 6);
    CholeskyDecomposition(lowermat, tmpsqr, 6, 6);
/*
    DumpMatrix(L"Matrix Dump\\Jacobian.txt", jacobian, noofpairs, 6);
    DumpMatrix(L"Matrix Dump\\TempSqr.txt", tmpsqr, 6, 6);
    DumpMatrix(L"Matrix Dump\\LowerMat.txt", lowermat, 6, 6);
    */

    MatMultTA(b, jacobian, noofpairs, 6, r, noofpairs);
    for (i=0; i<6; i++)
    {
      int ti=0;
      double sum=b[i];
      for (j=0; j<i; j++, ti+=6)
        sum-=lowermat[i+ti]*y[j];
      y[i]=sum/lowermat[i+ti];
    }

    for (i=5; i>=0; --i)
    {
      double sum=y[i];
      int ti=i*6;
      for (j=i+1; j<6; j++)
        sum-=lowermat[j+ti]*update[j];
      update[i]=sum/lowermat[ti+i];
    }

//    DumpMatrix(L"Matrix Dump\\Updates.txt", update, 6, 1);

    for (i=0; i<6; i++)
      variables[i]-=update[i];

    CalcCosSins(coss, sins, variables);
    CalcTransform(transform, coss, sins, variables+3);

    double error=CalcInnerICPError();

    double diff=besterror-error;

    if (diff>ITERATION_PRECISION)//(error<besterror)
    {
      for (i=0; i<3; i++)
      {
        cameratrack.rxyz[i]=variables[i];
        cameratrack.txyz[i]=variables[3+i];
      }
      besterror=error;
    }
    else if (iteration<MIN_ITERATIONS)
    {
      double scale=1.0-((double) (MIN_ITERATIONS-iteration)/MIN_ITERATIONS);
      for (i=0; i<6; i++)
        variables[i]+=update[i]*scale;

      CalcCosSins(coss, sins, variables);
      CalcTransform(transform, coss, sins, variables+3);
    }
    else
      iteration=MAX_ITERATIONS;

    wchar_t msg[MAX_PATH];
    DWORD now=::GetTickCount();

    wcscpy_s(msg, MAX_PATH, L"    Sub iteration ");
    int len=(int) wcslen(msg);
    _itow_s(iteration, msg+len, MAX_PATH-len, 10);

    wcscat_s(msg, MAX_PATH, L": ");
    len+=wcslen(msg+len);
    _itow_s(now-start, msg+len, MAX_PATH-len, 10);
    wcscat_s(msg, MAX_PATH, L" ms - Error:");

    len+=(int) wcslen(msg+len);
    FormatNumber(msg+len, MAX_PATH-len, error, 8);
    wcscat_s(msg, MAX_PATH, L"\r\n");
    AppendInfoText(msg, 0);

    start=now;

  } while (++iteration<MAX_ITERATIONS);

//  CalcCosSins(cameratrack.coss, cameratrack.sins, cameratrack.rxyz);
}

double MainWnd::TrackDataOuter(POINTDATA* frame, const int used)
{
  int i, iteration=0;
  double besterror=-1;
  double transform[16];

  do
  {
    wchar_t msg[MAX_PATH];
    GetTime(msg, MAX_PATH);
    wcscat_s(msg, MAX_PATH, L"Outer Iteration ");
    int len=(int) wcslen(msg);
    _itow_s(iteration, msg+len, MAX_PATH-len, 10);
    AppendInfoText(msg, 0);

    CalcCosSins(cameratrack.coss, cameratrack.sins, cameratrack.rxyz);
    CalcTransform(transform, cameratrack.coss, cameratrack.sins, cameratrack.txyz);

    DWORD start=::GetTickCount();

    pairmatch.FindPairsV2(frame, transform, &clouddata, pairratio[0]>50);//true);
//    SavePairs(frame, DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT, NULL, 0, transform, true);

    int noofpairs=0;
    int noofsubpairs=0;
    double sumerror=0;
    for (i=0; i<DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT; i++)
    {
      if (/*pointdata[i].used && */pointdata[i].matchpoint)
      {
        sumerror+=pointdata[i].matchdist;
        if (pointdata[i].matchdist<INIT_DISTANCE_THRESHOLD)
          ++noofsubpairs;
        
        pointdata[i].matchpoint->paired=true;
        pointdata[i].paired=true;
        
        pairs[noofpairs].pt1=pointdata+i;
        pairs[noofpairs++].pt2=pointdata[i].matchpoint;
//          pairs[noofpairs++].error=pointdata[i].matchdist;
      }
    }
    if (noofpairs)
      sumerror/=noofpairs;

    assert(noofpairs<=DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT);

    wcscpy_s(msg, MAX_PATH, L": Error: ");
    len=(int) wcslen(msg);
    FormatNumber(msg+len, MAX_PATH-len, sumerror, 5);
    wcscat_s(msg, MAX_PATH, L": DT: ");
    len=(int) wcslen(msg);
    FormatNumber(msg+len, MAX_PATH-len, DISTANCE_THRESHOLD, 3);
    wcscat_s(msg, MAX_PATH, L"\r\n");
    AppendInfoText(msg, 0);

    msg[0]=msg[1]=' ';
    GetTime(msg+2, MAX_PATH-2);
    wcscat_s(msg, MAX_PATH, L"Finding pairs ");
    AppendInfoText(msg, 0);

    pairratio[0]=(100.0f*noofpairs)/used;
    pairratio[1]=(100.0f*noofsubpairs)/noofpairs;
/*    if (pairratio>90)
    {
      if (DISTANCE_THRESHOLD>0.055)
        DISTANCE_THRESHOLD-=0.01f;
//        DISTANCE_THRESHOLD=0.05f;
    }
    else if (DISTANCE_THRESHOLD<1.0f)
      DISTANCE_THRESHOLD+=0.1f;*/

    DWORD now=::GetTickCount();
    msg[0]='(';
    FormatNumber(msg+1, MAX_PATH-1, pairratio[0], 2);
//      _itow_s(noofpairs, msg+1, MAX_PATH-1, 10);
    wcscat_s(msg, MAX_PATH, L"/");
    len=(int) wcslen(msg);
    FormatNumber(msg+len, MAX_PATH-len, pairratio[1], 2);
    wcscat_s(msg, MAX_PATH, L"%) - ");
    len=wcslen(msg);
    _itow_s(used, msg+len, MAX_PATH-len, 10);
    wcscat_s(msg, MAX_PATH, L" in ");
    len=wcslen(msg);
    _itow_s(now-start, msg+len, MAX_PATH-len, 10);
    wcscat_s(msg, MAX_PATH, L"ms)\r\n");
    AppendInfoText(msg, 0);

    if (!noofpairs || (iteration && (besterror-sumerror)<ITERATION_PRECISION))
      iteration=MAX_ITERATIONS;
    else
    {
      besterror=sumerror;
      TrackDataInner(frame, used, noofpairs);
    }

  } while (++iteration<MAX_ITERATIONS);

  return besterror;
}


void MainWnd::TrackDataAux(POINTDATA* frame, const int used)
{
  if (used>6)
  {
    int i;
    wchar_t msg[MAX_PATH];
    GetTime(msg, MAX_PATH);
    wcscat_s(msg, MAX_PATH, L"Frame #");
    int len=(int) wcslen(msg);
    _itow_s(trackedframes, msg+len, MAX_PATH-len, 10);
    wcscat_s(msg, MAX_PATH, L" (");
    len=(int) wcslen(msg);
    _itow_s(noofwldpts, msg+len, MAX_PATH-len, 10);
    wcscat_s(msg, MAX_PATH, L")\r\n");

//    wcscat_s(msg, MAX_PATH, L"Tracking Frame\r\n");
    len+=(int) wcslen(msg+len);
    WriteCameraTrackInfo(msg+len, MAX_PATH-len);
    wcscat_s(msg, MAX_PATH, L"\r\n");
    AppendInfoText(msg, 1);

    double error;
    int iteration=0;

    DISTANCE_THRESHOLD=INIT_DISTANCE_THRESHOLD;

    do
    {
      error=TrackDataOuter(frame, used);

      if (pairratio[0]<90 && DISTANCE_THRESHOLD<1.0f)
        DISTANCE_THRESHOLD+=0.1f;
      else
        iteration=10;
/*      if (pairratio>90)
      {
        if (DISTANCE_THRESHOLD>0.055)
          DISTANCE_THRESHOLD-=0.01f;
  //        DISTANCE_THRESHOLD=0.05f;
      }
      else if (DISTANCE_THRESHOLD<1.0f)
        DISTANCE_THRESHOLD+=0.1f;*/

    } while (pairratio[0]<90 && ++iteration<10);

    double transform[16];
    CalcCosSins(cameratrack.coss, cameratrack.sins, cameratrack.rxyz);
    CalcTransform(transform, cameratrack.coss, cameratrack.sins, cameratrack.txyz);

    pairmatch.FindPairsV2(frame, transform, &clouddata, pairratio[0]>50);//true);
    SavePairs(frame, DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT, NULL, 0, transform, true);


//    CalcCosSins(cameratrack.coss, cameratrack.sins, cameratrack.rxyz);
//    CalcTransform(transform, cameratrack.coss, cameratrack.sins, cameratrack.txyz);


    //if (grabdata || (/*error<10 && */lastgrab>10))
/*    if (error>(DISTANCE_THRESHOLD/2.0f) || true)
      --trackedframes;
    else */if (pairratio[0]>75 && pairratio[1]>75)
    {
      lastgrab=trackedframes;
      renderer[1].SetPointData(NULL, 0, false);

      grabdata=false;

      msg[0]=msg[1]=' ';
      GetTime(msg+2, MAX_PATH-2);
      wcscat_s(msg, MAX_PATH, L"Grab: ");
      AppendInfoText(msg, 0);
      DWORD start=::GetTickCount();

      int counts[2];

      int noofinvpts=0;
      double transform[16], inverse[16];
      float min[4], max[4], f_near[4], f_far[4];

      CalcTransform(transform, cameratrack.coss, cameratrack.sins, cameratrack.txyz);
      CalcInverseTransform(inverse, cameratrack.coss, cameratrack.sins, cameratrack.txyz);
      CalcFrameMinMax(transform, min, max);
//      POINTDATA** invpts=clouddata.GetPoints(min, max, noofinvpts, 0, true);

      CalcFrustum(f_near, f_far);

      POINTDATA** invpts=clouddata.GetPoints(min, max, inverse, f_near, f_far, noofinvpts, 0, false);

      f_near[0]-=REGION_SIZE;
      f_near[1]-=REGION_SIZE;
      f_near[2]-=REGION_SIZE;

      f_far[0]-=REGION_SIZE;
      f_far[1]-=REGION_SIZE;
      f_far[2]+=REGION_SIZE;

      const float rangex=f_far[0]-f_near[0];
      const float rangey=f_far[1]-f_near[1];
      const float rangez=f_near[2]-f_far[2];
/*
      for (i=0; i<DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT; i++)
      {
        if (frame[i].used)// && !frame[i].matchpoint)
        {
          float position=f_near[2]-frame[i].xyz[2];
          float scale=position/rangez;
          assert(scale>=0.0f && scale<=1.0f);

          float tx=f_near[0]+rangex*scale;
          float ty=f_near[1]+rangey*scale;

          if (frame[i].xyz[0]<-tx || frame[i].xyz[0]>tx || frame[i].xyz[1]<-ty || frame[i].xyz[1]>ty)
          {
              frame[i].used=false;
//              frame[i].matchpoint=NULL;
          }
        }
      }*/

      pairmatch.FindInversePairs(invpts, noofinvpts, frame, inverse, false);

      //SavePairs(frame, DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT, invpts, noofinvpts, transform, false);

//      worlddata=pdupdate.UpdatePointCloud(frame, transform, counts, worlddata, noofwldpts, maxwldpts, trackedframes, false);
/*
      for (i=0; i<3; i++)
      {
        min_l[i]+=10*MATCH_DISTANCE;
        max_l[i]-=10*MATCH_DISTANCE;
      }*/

      counts[0]=counts[1]=0;
      for (i=0; i<noofinvpts; i++)
      {
        if (invpts[i]->matchpoint)
        {
          int tc=invpts[i]->count+1;
          float nxyz[3];
          POINTDATA* pt=invpts[i]->matchpoint;
          
          nxyz[0]=(float) (transform[0]*pt->nxyz[0]+transform[4]*pt->nxyz[1]+transform[8]*pt->nxyz[2]);
          nxyz[1]=(float) (transform[1]*pt->nxyz[0]+transform[5]*pt->nxyz[1]+transform[9]*pt->nxyz[2]);
          nxyz[2]=(float) (transform[2]*pt->nxyz[0]+transform[6]*pt->nxyz[1]+transform[10]*pt->nxyz[2]);

          float tmp=0;
          for (int j=0; j<3; j++)
          {
            // if the position is updated then need to update the clouddata
  //          matchpt->xyz[j]=(float) ((txyz[j]+matchpt->xyz[j]*matchpt->count)/tc);
            invpts[i]->nxyz[j]=(float) ((nxyz[j]+invpts[i]->nxyz[j]*invpts[i]->count)/tc);
            invpts[i]->colour[j]=(float) ((pt->colour[j]+invpts[i]->colour[j]*invpts[i]->count)/tc);
            tmp+=invpts[i]->nxyz[j]*invpts[i]->nxyz[j];
          }
          if (tmp>0)
          {
            tmp=sqrt(tmp);
            invpts[i]->nxyz[0]/=tmp;
            invpts[i]->nxyz[1]/=tmp;
            invpts[i]->nxyz[2]/=tmp;
          }
          else
          {
            invpts[i]->nxyz[0]=0.57735f;
            invpts[i]->nxyz[1]=0.57735f;
            invpts[i]->nxyz[2]=0.57735f;
          }

          invpts[i]->lastseen=trackedframes;
          if (tc<256)
            ++invpts[i]->count;

          ++counts[1];
        }
        else
        {
          // prune
          float tmp1[4], tmp2[4];
          tmp1[0]=invpts[i]->xyz[0];
          tmp1[1]=invpts[i]->xyz[1];
          tmp1[2]=invpts[i]->xyz[2];
          tmp1[3]=1.0f;
          MatMult(tmp2, inverse, 4, 4, tmp1, 4, 1);

          if (tmp2[2]<f_near[2] && tmp2[2]>f_far[2])
          {
            float position=f_near[2]-tmp2[2];
            float scale=position/rangez;
            assert(scale>=0.0f && scale<=1.0f);

            float tx=f_near[0]+rangex*scale;
            float ty=f_near[1]+rangey*scale;

            if (tmp2[0]>-tx && tmp2[0]<tx && tmp2[1]>-ty && tmp2[1]<ty)
              clouddata.CheckPrune(worlddata, noofwldpts, invpts[i], trackedframes);
          }
        }
      }
      
      for (i=0; i<DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT; i++)
      {
        if (frame[i].used && !frame[i].matchpoint && frame[i].xyz[2]<f_near[2] && frame[i].xyz[2]>f_far[2])
        {
          float position=f_near[2]-frame[i].xyz[2];
          float scale=position/rangez;
          assert(scale>=0.0f && scale<=1.0f);

          float tx=f_near[0]+rangex*scale;
          float ty=f_near[1]+rangey*scale;

          if (frame[i].xyz[0]>-tx && frame[i].xyz[0]<tx && frame[i].xyz[1]>-ty && frame[i].xyz[1]<ty)
          {
            if (noofwldpts==maxwldpts)
            {
              maxwldpts+=100;
              POINTDATA* tmp=worlddata;
              worlddata=(POINTDATA*) realloc(worlddata, sizeof(POINTDATA)*maxwldpts);
              if (tmp!=worlddata)
              {
                clouddata.ClearPointData();
                clouddata.AddPointData(worlddata, noofwldpts);
              }
            }

            POINTDATA* pt=frame+i;
            worlddata[noofwldpts]=frame[i];
        
            worlddata[noofwldpts].xyz[0]=(float) (transform[0]*pt->xyz[0]+transform[4]*pt->xyz[1]+transform[8]*pt->xyz[2]+transform[12]);
            worlddata[noofwldpts].xyz[1]=(float) (transform[1]*pt->xyz[0]+transform[5]*pt->xyz[1]+transform[9]*pt->xyz[2]+transform[13]);
            worlddata[noofwldpts].xyz[2]=(float) (transform[2]*pt->xyz[0]+transform[6]*pt->xyz[1]+transform[10]*pt->xyz[2]+transform[14]);

            worlddata[noofwldpts].nxyz[0]=(float) (transform[0]*pt->nxyz[0]+transform[4]*pt->nxyz[1]+transform[8]*pt->nxyz[2]);
            worlddata[noofwldpts].nxyz[1]=(float) (transform[1]*pt->nxyz[0]+transform[5]*pt->nxyz[1]+transform[9]*pt->nxyz[2]);
            worlddata[noofwldpts].nxyz[2]=(float) (transform[2]*pt->nxyz[0]+transform[6]*pt->nxyz[1]+transform[10]*pt->nxyz[2]);

            worlddata[noofwldpts].paired=false;
            worlddata[noofwldpts].used=true;
            worlddata[noofwldpts].lastseen=trackedframes;
            worlddata[noofwldpts].count=1;

            clouddata.AddPointData(worlddata+noofwldpts);
            ++noofwldpts;
            ++counts[0];
          }
        }
      }

      _itow_s(counts[0], msg, MAX_PATH, 10);
      wcscat_s(msg, MAX_PATH, L" new; ");

      int len=(int) wcslen(msg);
      _itow_s(counts[1], msg+len, MAX_PATH-len, 10);
      wcscat_s(msg, MAX_PATH, L" updated in ");

      len+=wcslen(msg+len);
      _itow_s(::GetTickCount()-start, msg+len, MAX_PATH-len, 10);
      wcscat_s(msg, MAX_PATH, L" ms\r\n");

      AppendInfoText(msg, 0);
/*
      if (trackedframes-lastpruned>DROP_RATE/4)
      {
        float tmpvec[4], min[4], max[4];
        tmpvec[0]=pointdata[0].xyz[0];
        tmpvec[1]=pointdata[0].xyz[1];
        tmpvec[2]=CalcDataDepth(depthrange[1]);
        tmpvec[3]=1.0f;
        MatMult(min, transform, 4, 4, tmpvec, 4, 1);

        tmpvec[0]=-pointdata[0].xyz[0];
        tmpvec[1]=-pointdata[0].xyz[1];
        tmpvec[2]=CalcDataDepth(depthrange[0]);
        MatMult(max, transform, 4, 4, tmpvec, 4, 1);

        clouddata.Prune(worlddata, noofwldpts, min, max, trackedframes);
        lastpruned=trackedframes;
      }*/

      lastpruned=trackedframes;
      renderer[1].SetPointData(worlddata, noofwldpts, false);
    }

//    TrackData();    // turns off tracking so 1 frame at a time
    OnGLRedraw(2);
  }
}

void MainWnd::InitICPErrorThreadMemory(const double* transform, const int noofpairs, const PAIR* pairs)
{
  int i, offset=0;
  int ppt=noofpairs/noofthreads;
  int rem=noofpairs-(ppt*noofthreads);

  offset=ppt+rem;
  icpthreadinfo[0].pairs=pairs;
  icpthreadinfo[0].noofpairs=offset;
  icpthreadinfo[0].r=r;
  icpthreadinfo[0].transform=transform;

  for (i=1; i<noofthreads; i++)
  {
    icpthreadinfo[i].pairs=pairs+offset;
    icpthreadinfo[i].noofpairs=ppt;
    icpthreadinfo[i].r=r+offset;
    icpthreadinfo[i].transform=transform;

    offset+=ppt;
  }
}


void MainWnd::InitJacobianThreadMemory(const double* coss, const double* sins, const int noofpairs, const PAIR* pairs)
{
  int i, offset=0;
  int ppt=noofpairs/noofthreads;
  int rem=noofpairs-(ppt*noofthreads);

  offset=ppt+rem;
  jacobthreadinfo[0].offset=0;
  jacobthreadinfo[0].pairs=pairs;
  jacobthreadinfo[0].noofpairs=offset;
  jacobthreadinfo[0].totalpairs=noofpairs;
  jacobthreadinfo[0].coss=coss;
  jacobthreadinfo[0].sins=sins;

  for (i=1; i<noofthreads; i++)
  {
    jacobthreadinfo[i].offset=offset;
    jacobthreadinfo[i].pairs=pairs+offset;
    jacobthreadinfo[i].noofpairs=ppt;
    jacobthreadinfo[i].totalpairs=noofpairs;
    jacobthreadinfo[i].coss=coss;
    jacobthreadinfo[i].sins=sins;

    offset+=ppt;
  }
}

void MainWnd::CalcJacobian()
{
  for (int i=0; i<noofthreads; i++)
  {
    ResetEvent(jacobthreadinfo[i].alerts[1]);
    SetEvent(jacobthreadinfo[i].alerts[0]);
  }

  WaitForMultipleObjects(noofthreads, alerts, TRUE, INFINITE);
}

DWORD WINAPI JacobianThread(LPVOID lpParameter)
{
  const JACOBIANTHREADINFO* info=(JACOBIANTHREADINFO*) lpParameter;

  while (*info->alive)
  {
    WaitForSingleObject(info->alerts[0], INFINITE);
    ResetEvent(info->alerts[0]);
    
    if (*info->alive)
    {
      const double* coss=info->coss;
      const double* sins=info->sins;
      const int noofpairs=info->noofpairs;

      const double t6=coss[0]*coss[2];//t1*t3;
      const double t7=sins[0]*sins[2];//t4*t5;
      const double t8=t6*sins[1]-t7;//t6*t2-t7;
      const double t9=coss[0]*sins[2];//t1*t5;
      const double t10=sins[0]*coss[2];//t4*t3;
      const double t11=-t9*sins[1]-t10;//-t9*t2-t10;
      const double t9_2=t10*sins[1]+t9;//t10*t2+t9;
      const double t6_2=t6-t7*sins[1];//-t7*t2+t6;
    //      t10=t7*coss[1];//t7*t12;

      for (int i=0; i<noofpairs; i++)
      {
        POINTDATA* source=info->pairs[i].pt1;
        POINTDATA* target=info->pairs[i].pt2;

        const double t13=coss[1]*source->xyz[2];//t12*Vz;
/*        t9=t10*sins[1]+t9;//t10*t2+t9;
        t6=t6-t7*sins[1];//-t7*t2+t6;*/
        const double t7_2=sins[2]*source->xyz[1]-coss[2]*source->xyz[0];//-t3*Vx+t5*Vy;
        const double t10_2=t7_2*coss[1];//t7*t12;
        const double t14=sins[1]*source->xyz[2];//t2*Vz;

        const double nxw=source->matchweight*target->nxyz[0];
        const double nyw=source->matchweight*target->nxyz[1];
        const double nzw=source->matchweight*target->nxyz[2];
      
        int index=info->offset+i;
/*        info->jacobian[index]=(t8*source->xyz[0]+t11*source->xyz[1]-t13*coss[0])*target->nxyz[1]+(t9_2*source->xyz[0]+t6_2*source->xyz[1]-t13*sins[0])*target->nxyz[2];//(t8 * Vx + t11 * Vy - t13 * t1) * Ny + (t9 * Vx + t6 * Vy - t13 * t4) * Nz;
        info->jacobian[index+=info->totalpairs]=(t7_2*sins[1]+t13)*target->nxyz[0]+sins[0]*(t14-t10_2)*target->nxyz[1]+coss[0]*(t10_2-t14)*target->nxyz[2];//(t7 * t2 + t13) * Nx + t4 * (-t10 + t14) * Ny + t1 * (t10 - t14) * Nz;
        info->jacobian[index+=info->totalpairs]=(t6_2*source->xyz[0]-t9_2*source->xyz[1])*target->nxyz[1]+(t8*source->xyz[1]-t11*source->xyz[0])*target->nxyz[2]-coss[1]*(sins[2]*source->xyz[0]+coss[2]*source->xyz[1])*target->nxyz[0];//-t12 * (t5 * Vx + t3 * Vy) * Nx + (t6 * Vx - t9 * Vy) * Ny + (-t11 * Vx + t8 * Vy) * Nz;
        info->jacobian[index+=info->totalpairs]=target->nxyz[0];//Nx;
        info->jacobian[index+=info->totalpairs]=target->nxyz[1];//Ny;
        info->jacobian[index+info->totalpairs]=target->nxyz[2];//Nz;*/

        info->jacobian[index]=(t8*source->xyz[0]+t11*source->xyz[1]-t13*coss[0])*nyw+(t9_2*source->xyz[0]+t6_2*source->xyz[1]-t13*sins[0])*nzw;//(t8 * Vx + t11 * Vy - t13 * t1) * Ny + (t9 * Vx + t6 * Vy - t13 * t4) * Nz;
        info->jacobian[index+=info->totalpairs]=(t7_2*sins[1]+t13)*nxw+sins[0]*(t14-t10_2)*nyw+coss[0]*(t10_2-t14)*nzw;//(t7 * t2 + t13) * Nx + t4 * (-t10 + t14) * Ny + t1 * (t10 - t14) * Nz;
        info->jacobian[index+=info->totalpairs]=(t6_2*source->xyz[0]-t9_2*source->xyz[1])*nyw+(t8*source->xyz[1]-t11*source->xyz[0])*nzw-coss[1]*(sins[2]*source->xyz[0]+coss[2]*source->xyz[1])*nxw;//-t12 * (t5 * Vx + t3 * Vy) * Nx + (t6 * Vx - t9 * Vy) * Ny + (-t11 * Vx + t8 * Vy) * Nz;
        info->jacobian[index+=info->totalpairs]=nxw;//Nx;
        info->jacobian[index+=info->totalpairs]=nyw;//Ny;
        info->jacobian[index+info->totalpairs]=nzw;//Nz;
      }
    }
    SetEvent(info->alerts[1]);
  }

  return 0;
}

double MainWnd::CalcInnerICPError()//const PAIR* pairs, const int noofpairs, const float* transform)
{
  for (int i=0; i<noofthreads; i++)
  {
    ResetEvent(icpthreadinfo[i].alerts[1]);
    SetEvent(icpthreadinfo[i].alerts[0]);
  }

  WaitForMultipleObjects(noofthreads, alerts, TRUE, INFINITE);

  double error=0;
  for (int i=0; i<noofthreads; i++)
    error+=icpthreadinfo[i].error;
  return error;
}

DWORD WINAPI ICPErrorThread(LPVOID lpParameter)
{
  ICPERRORTHREADINFO* info=(ICPERRORTHREADINFO*) lpParameter;

  while (*info->alive)
  {
    WaitForSingleObject(info->alerts[0], INFINITE);
    ResetEvent(info->alerts[0]);
    
    if (*info->alive)
    {
      double error=0;
      const int noofpairs=info->noofpairs;
      const PAIR* pairs=info->pairs;
      const double* transform=info->transform;

      for (int i=0; i<noofpairs; i++)
      {
        POINTDATA* source=pairs[i].pt1;//pointdata+pairs[i].id1;
        POINTDATA* target=pairs[i].pt2;//worlddata+pairs[i].id2;

        double tx=transform[0]*source->xyz[0]+transform[4]*source->xyz[1]+transform[8]*source->xyz[2]+transform[12]-target->xyz[0];
        double ty=transform[1]*source->xyz[0]+transform[5]*source->xyz[1]+transform[9]*source->xyz[2]+transform[13]-target->xyz[1];
        double tz=transform[2]*source->xyz[0]+transform[6]*source->xyz[1]+transform[10]*source->xyz[2]+transform[14]-target->xyz[2];

        double r=tx*target->nxyz[0]+ty*target->nxyz[1]+tz*target->nxyz[2];
        r*=source->matchweight;
        error+=r*r;
        info->r[i]=r;
      }

      info->error=error;
    }
    SetEvent(info->alerts[1]);
  }

  return 0;
}

void MainWnd::SavePairs(const POINTDATA* frame, const int framesize, POINTDATA** wldpts, const int noofwldpts, const double* transform, const bool toworld)
{
  FILE* file=NULL;
  const float Z_SHIFT=(depthrange[1]/2000.0f);

  if (!_wfopen_s(&file, MATCH_FILE, L"a+b") && file)
  {
    unsigned char tmp=(toworld?0:1);
    fwrite(&trackedframes, sizeof(int), 1, file);
    fwrite(&tmp, 1, 1, file);

    if (wldpts)
    {
      fwrite(&noofwldpts, sizeof(int), 1, file);
      for (int i=0; i<noofwldpts; i++)
      {
        float xyz[3];
        memcpy(xyz, wldpts[i]->xyz, 3*sizeof(float));
        xyz[2]+=Z_SHIFT;
        fwrite(xyz, sizeof(float), 3, file);
        fwrite(wldpts[i]->nxyz, sizeof(float), 3, file);

        if (!toworld)
        {
          int offset=(wldpts[i]->matchpoint?wldpts[i]->matchpoint-frame:-1);
          fwrite(&offset, sizeof(int), 1, file);
        }
      }
    }
    else
    {
      fwrite(&this->noofwldpts, sizeof(int), 1, file);
      for (int i=0; i<this->noofwldpts; i++)
      {
        float xyz[3];
        memcpy(xyz, worlddata[i].xyz, 3*sizeof(float));
        xyz[2]+=Z_SHIFT;
        fwrite(xyz, sizeof(float), 3, file);
        fwrite(worlddata[i].nxyz, sizeof(float), 3, file);

        if (!toworld)
        {
          int offset=(worlddata[i].matchpoint?worlddata[i].matchpoint-frame:-1);
          fwrite(&offset, sizeof(int), 1, file);
        }
      }
    }

    int count=0;
    for (int i=0; i<framesize; i++)
      if (frame[i].used)
        ++count;

    fwrite(&count, sizeof(int), 1, file);
    for (int i=0; i<framesize; i++)
    {
      if (frame[i].used)
      {
        float txyz[3], nxyz[3];
        const POINTDATA* pt=frame+i;
        txyz[0]=(float) (transform[0]*pt->xyz[0]+transform[4]*pt->xyz[1]+transform[8]*pt->xyz[2]+transform[12]);
        txyz[1]=(float) (transform[1]*pt->xyz[0]+transform[5]*pt->xyz[1]+transform[9]*pt->xyz[2]+transform[13]);
        txyz[2]=(float) (transform[2]*pt->xyz[0]+transform[6]*pt->xyz[1]+transform[10]*pt->xyz[2]+transform[14]);

        nxyz[0]=(float) (transform[0]*pt->nxyz[0]+transform[4]*pt->nxyz[1]+transform[8]*pt->nxyz[2]);
        nxyz[1]=(float) (transform[1]*pt->nxyz[0]+transform[5]*pt->nxyz[1]+transform[9]*pt->nxyz[2]);
        nxyz[2]=(float) (transform[2]*pt->nxyz[0]+transform[6]*pt->nxyz[1]+transform[10]*pt->nxyz[2]);
            
        txyz[2]+=Z_SHIFT;

        fwrite(txyz, sizeof(float), 3, file);
        fwrite(nxyz, sizeof(float), 3, file);

        if (toworld)
        {
          int offset=(pt->matchpoint?pt->matchpoint-worlddata:-1);
          fwrite(&offset, sizeof(int), 1, file);
        }
        else
        {
          int tmp=pt->matchpoint?1:-1;
          fwrite(&tmp, sizeof(int), 1, file);
          fwrite(&i, sizeof(int), 1, file);
        }
      }
    }
    fclose(file);
  }
}

void MainWnd::ScreenGrab()
{
  RECT rect;
  HDC hdcsrc=::GetDC(NULL);
  ::GetWindowRect(hwnd, &rect);

  int width=rect.right-rect.left;
  int height=rect.bottom-rect.top;

  HDC hdctmp=::CreateCompatibleDC(hdcsrc);

  BITMAPINFO bminfo;
  memset(&bminfo, 0, sizeof(BITMAPINFO));
  bminfo.bmiHeader.biSize=sizeof(BITMAPINFO);
  bminfo.bmiHeader.biWidth=width;
  bminfo.bmiHeader.biHeight=height;
  bminfo.bmiHeader.biPlanes=1;
  bminfo.bmiHeader.biBitCount=32;
  bminfo.bmiHeader.biCompression=BI_RGB;
  bminfo.bmiHeader.biSizeImage=bminfo.bmiHeader.biWidth*bminfo.bmiHeader.biHeight*bminfo.bmiHeader.biBitCount/8;
  
  DWORD* bmpdata=NULL;
  HBITMAP hbmp=::CreateDIBSection(hdcsrc, &bminfo, DIB_RGB_COLORS, (void**) &bmpdata, NULL, NULL);
  ::SelectObject(hdctmp, hbmp);
  ::BitBlt(hdctmp, 0, 0, width, height, hdcsrc, rect.left, rect.top, SRCCOPY);
  ::ReleaseDC(NULL, hdcsrc);


  wchar_t filename[MAX_PATH];
  wcscpy_s(filename, MAX_PATH, SCREEN_GRAB_FOLDER);

  wchar_t* tmp=wcsrchr(filename, '\\');
  tmp[0]='\0';
  CreateDirectory(filename, NULL);
  tmp[0]='\\';

  wcscat_s(filename, MAX_PATH, L"Output");
  int len=(int) wcslen(filename);

  int frame=trackedframes;
  do
  {
    int sublen=len;
    if (frame<1000)
      filename[sublen++]='0';
    if (frame<100)
      filename[sublen++]='0';
    if (frame<10)
      filename[sublen++]='0';
    _itow_s(frame, filename+sublen, MAX_PATH-sublen, 10);
    wcscat_s(filename, MAX_PATH, L".bmp");
    
    ++frame;
  } while (GetFileAttributes(filename)!=INVALID_FILE_ATTRIBUTES);
  

  BITMAPINFO bmpinfo;
  memset(&bmpinfo, 0, sizeof(BITMAPINFO));
  bmpinfo.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
  bmpinfo.bmiHeader.biWidth=width;
  bmpinfo.bmiHeader.biHeight=height;
  bmpinfo.bmiHeader.biPlanes=1;
  bmpinfo.bmiHeader.biBitCount=24;
  bmpinfo.bmiHeader.biCompression=BI_RGB;
  bmpinfo.bmiHeader.biSizeImage=bmpinfo.bmiHeader.biWidth*bmpinfo.bmiHeader.biHeight*bmpinfo.bmiHeader.biBitCount/8;
  bmpinfo.bmiHeader.biXPelsPerMeter=0;
  bmpinfo.bmiHeader.biYPelsPerMeter=0;
  bmpinfo.bmiHeader.biClrUsed=0;
  bmpinfo.bmiHeader.biClrImportant=0;

  BITMAPFILEHEADER fheader;
  fheader.bfType=0x4D42;
  fheader.bfReserved1=fheader.bfReserved2=0;
  fheader.bfOffBits=sizeof(BITMAPINFO)+sizeof(BITMAPFILEHEADER);
  fheader.bfSize=sizeof(BITMAPINFO)+sizeof(BITMAPFILEHEADER)+bmpinfo.bmiHeader.biSizeImage;

  FILE* file;
  _wfopen_s(&file, filename, L"wb");
  fwrite(&fheader, sizeof(BITMAPFILEHEADER), 1, file);
  fwrite(&bmpinfo, sizeof(BITMAPINFO), 1, file);
  for (int j=0; j<height; j++)
  {
    int pos=j*width;
    for (int i=0; i<width; i++)
    {
      BYTE* data=(BYTE*) (bmpdata+pos);
      fwrite(data, 1, 1, file);
      fwrite(data+1, 1, 1, file);
      fwrite(data+2, 1, 1, file);
      ++pos;
    }
  }
  fclose(file);

  ::DeleteObject(hbmp);
  ::DeleteDC(hdctmp);
}

void MainWnd::ExportWRLModel(bool frame)
{
  wchar_t filename[MAX_PATH];
  wcscpy_s(filename, MAX_PATH, MODEL_OUTPUT);
  
  int i, j, noofpts;
  POINTDATA* data;
  
  if (frame)
  {
    data=pointdata;
    noofpts=DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT;
    wcscat_s(filename, MAX_PATH, L"Frame.wrl");
    exportdata[0]=false;
  }
  else
  {
    data=worlddata;
    noofpts=noofwldpts;
    for (i=0; i<noofpts; i++)
      data[i].used=true;

    wcscat_s(filename, MAX_PATH, L"World.wrl");
    exportdata[1]=false;
  }

  char cbuff[100];
  wchar_t wbuff[100];
/*
  // each as cubes
  FILE* file=NULL;
  if (!_wfopen_s(&file, filename, L"wb") && file)
  {
    fwrite("#VRML V1.0 ascii\r\n\r\n", 1, 20, file);
    for (i=0; i<noofpts; i++)
    {
      if (data[i].used)
      {
        fwrite("Separator\r\n{\r\n", 1, 14, file);
      
        fwrite("  Translation { translation ", 1, 28, file);
        for (j=0; j<3; j++)
        {
          FormatNumber(wbuff, 100, data[i].xyz[j], 5);
          WideCharToMultiByte(CP_ACP, 0, wbuff, -1, cbuff, 100, 0, NULL);
          fwrite(cbuff, 1, strlen(cbuff), file);
          fwrite(" ", 1, 1, file);
        }
        fwrite("}\r\n", 1, 3, file);

        fwrite("  Material { diffuseColor	", 1, 26, file);
        for (j=0; j<3; j++)
        {
          FormatNumber(wbuff, 100, data[i].colour[j], 5);
          WideCharToMultiByte(CP_ACP, 0, wbuff, -1, cbuff, 100, 0, NULL);
          fwrite(cbuff, 1, strlen(cbuff), file);
          fwrite(" ", 1, 1, file);
        }
        fwrite("}\r\n", 1, 3, file);

        fwrite("  Cube { width 0.02 height 0.02 depth 0.02 }\r\n", 1, 46, file);

        fwrite("}\r\n\r\n", 1, 5, file);
      }
    }
    fclose(file);
  }*/

  // output as triangles
  float r=0.02f;
  float dr=(float) sqrt(0.5*r*r);
  FILE* file=NULL;
  if (!_wfopen_s(&file, filename, L"wb") && file)
  {
    fwrite("#VRML V1.0 ascii\r\n\r\n", 1, 20, file);
    fwrite("Separator\r\n{\r\n", 1, 14, file);
    fwrite("  Coordinate3\r\n  {\r\n    point\r\n    [", 1, 36, file);

    int count=0;
    for (i=0; i<noofpts; i++)
    {
      if (data[i].used)
      {
        if (count++)
          fwrite(",", 1, 1, file);
        
        fwrite("\r\n      ", 1, 8, file);

        float tmp[9];
        tmp[0]=data[i].xyz[0];
        tmp[2]=data[i].xyz[1]-r;
        tmp[1]=-data[i].xyz[2];

        tmp[3]=data[i].xyz[0]+dr;
        tmp[5]=data[i].xyz[1]+dr;
        tmp[4]=-data[i].xyz[2];

        tmp[6]=data[i].xyz[0]-dr;
        tmp[8]=data[i].xyz[1]+dr;
        tmp[7]=-data[i].xyz[2];

        for (j=0; j<9; j++)
        {
//          FormatNumber(wbuff, 100, data[i].xyz[j], 5);
          FormatNumber(wbuff, 100, tmp[j], 5);
          WideCharToMultiByte(CP_ACP, 0, wbuff, -1, cbuff, 100, 0, NULL);
          fwrite(cbuff, 1, strlen(cbuff), file);
          if (j==2 || j==5)
            fwrite(",\r\n      ", 1, 9, file);
          else if (j!=8)
            fwrite(" ", 1, 1, file);
        }
      }
    }
    fwrite("\r\n    ]\r\n  }", 1, 12, file);

    count=0;
    for (i=0; i<noofpts; i++)
    {
      if (data[i].used)
      {
        fwrite("\r\n\r\n  Material { diffuseColor	", 1, 30, file);
        for (j=0; j<3; j++)
        {
          FormatNumber(wbuff, 100, data[i].colour[j], 5);
          WideCharToMultiByte(CP_ACP, 0, wbuff, -1, cbuff, 100, 0, NULL);
          fwrite(cbuff, 1, strlen(cbuff), file);
          fwrite(" ", 1, 1, file);
        }
        fwrite("}\r\n", 1, 3, file);

        fwrite("	IndexedFaceSet { coordIndex [ ", 1, 31, file);
        for (j=0; j<3; j++)
        {
          _itoa_s(count++, cbuff, 100, 10);
          fwrite(cbuff, 1, strlen(cbuff), file);
          fwrite(", ", 1, 2, file);
        }
        fwrite("-1 ] }", 1, 6, file);
      }
    }

    fwrite("\r\n}", 1, 3, file);
    fclose(file);
  }
}

void MainWnd::CreatePolygonModel(bool frame)
{
  wchar_t filename[MAX_PATH];
  wcscpy_s(filename, MAX_PATH, POLYGON_TEMP_FILE);

  int i, noofpts;
  POINTDATA* data;
  
  if (frame)
  {
    int used=0;
    CalcNormals(pointdata, used);

    data=pointdata;
    noofpts=DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT;
    exportdata[0]=false;

    wcscat_s(filename, MAX_PATH, L"Frame");
  }
  else
  {
    data=worlddata;
    noofpts=noofwldpts;
    for (i=0; i<noofpts; i++)
      data[i].used=true;
    exportdata[1]=false;

    wcscat_s(filename, MAX_PATH, L"World");
  }

  wcscat_s(filename, MAX_PATH, L".dat");

  FILE* file=NULL;
  int count=0;
  for (i=0; i<noofpts; i++)
  {
    if (data[i].used)
      ++count;
  }

  if (!_wfopen_s(&file, filename, L"wb") && file)
  {
    fwrite(&count, 4, 1, file);
    for (i=0; i<noofpts; i++)
    {
      if (data[i].used)
      {
        fwrite(data[i].xyz, 4, 3, file);
        fwrite(data[i].nxyz, 4, 3, file);
      }
    }
    fclose(file);
  }

  ShellExecute(NULL, L"open", L"PCL_PCtoPoly.exe", filename, NULL , SW_SHOW);
}

DWORD WINAPI PlaybackThread(LPVOID lpParameter)
{
  MainWnd* parent=(MainWnd*) lpParameter;

  while (parent->playing && parent->recorder)
  {
    int used=parent->recorder->GetFrameData(parent->pointdata, DEPTH_FRAME_WIDTH, DEPTH_FRAME_HEIGHT);

    if (used)
    {
      parent->TrackData(parent->pointdata, used);
      parent->renderer[0].SetPointData(parent->pointdata, DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT, true);
    }
    else
      parent->Playback();
  }

  parent->recorder->CloseRecorder();
  delete parent->recorder;
  parent->recorder=NULL;

  return 0;
}
