#include "Matrices.h"
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

void MatMult(float* dest, const double* a, const int a_rows, const int a_cols, const float* b, const int b_rows, const int b_cols)
{
  assert(a_cols==b_rows);

  int pos=0;
  int boff=0;
  for (int j=0; j<b_cols; j++, boff+=b_rows)
  {
    for (int i=0; i<a_rows; i++)
    {
      int aoff=i;
      double sum=0;
      for (int k=0; k<a_cols; k++, aoff+=a_rows)
        sum+=a[aoff]*b[boff+k];
      dest[pos++]=(float) sum;
    }
  }
}

void MatMult(float* dest, const float* a, const int a_rows, const int a_cols, const float* b, const int b_rows, const int b_cols)
{
  assert(a_cols==b_rows);

  int pos=0;
  int boff=0;
  for (int j=0; j<b_cols; j++, boff+=b_rows)
  {
    for (int i=0; i<a_rows; i++)
    {
      int aoff=i;
      double sum=0;
      for (int k=0; k<a_cols; k++, aoff+=a_rows)
        sum+=a[aoff]*b[boff+k];
      dest[pos++]=(float) sum;
    }
  }
}

void MatMultTA_LT(double* dest, const double* a, const int a_rows, const int a_cols, const double* b, const int b_rows, const int b_cols)
{
  // this method only calculates the lower triangle of the multiplication including the diagonal
  // this does not zero the upper triangle of the dest matrix
  assert(a_rows==b_rows);

  int pos=0;
  int boff=0;
  for (int j=0; j<b_cols; j++, boff+=b_rows)
  {
    int aoff=j*a_rows;
    pos+=j;
    for (int i=j; i<a_cols; i++, aoff+=a_rows)
    {
      double sum=0;
      for (int k=0; k<a_rows; k++)
        sum+=a[aoff+k]*b[boff+k];
      dest[pos++]=sum;
    }
  }
}

void MatMultTA(double* dest, const double* a, const int a_rows, const int a_cols, const double* b, const int b_rows)
{
  assert(a_rows==b_rows);

  int pos=0;
  int aoff=0;
  for (int i=0; i<a_cols; i++, aoff+=a_rows)
  {
    double sum=0;
    for (int k=0; k<a_rows; k++)
      sum+=a[aoff+k]*b[k];
    dest[pos++]=sum;
  }
}

void CalcTransform(double* transform, const double* coss, const double* sins, const double* xyz)
{
  // first column
  transform[0]=coss[1]*coss[2];
  transform[1]=sins[0]*sins[1]*coss[2]+coss[0]*sins[2];
  transform[2]=-coss[0]*sins[1]*coss[2]+sins[0]*sins[2];
  transform[3]=0;

  // second column
  transform[4]=-coss[1]*sins[2];
  transform[5]=-sins[0]*sins[1]*sins[2]+coss[0]*coss[2];
  transform[6]=coss[0]*sins[1]*sins[2]+sins[0]*coss[2];
  transform[7]=0;

  // third column
  transform[8]=sins[1];
  transform[9]=-sins[0]*coss[1];
  transform[10]=coss[0]*coss[1];
  transform[11]=0;

  // forth column
  transform[12]=xyz[0];
  transform[13]=xyz[1];
  transform[14]=xyz[2];
  transform[15]=1;
}

void CalcInverseTransform(double* inverse, const double* coss, const double* sins, const double* xyz)
{
  // first column
  inverse[0]=coss[1]*coss[2];
  inverse[1]=-coss[1]*sins[2];
  inverse[2]=sins[1];
  inverse[3]=0;

  // second column
  inverse[4]=sins[0]*sins[1]*coss[2]+coss[0]*sins[2];
  inverse[5]=-sins[0]*sins[1]*sins[2]+coss[0]*coss[2];
  inverse[6]=-sins[0]*coss[1];
  inverse[7]=0;

  // third column
  inverse[8]=-coss[0]*sins[1]*coss[2]+sins[0]*sins[2];
  inverse[9]=coss[0]*sins[1]*sins[2]+sins[0]*coss[2];
  inverse[10]=coss[0]*coss[1];
  inverse[11]=0;

  // fourth column
  inverse[12]=-(inverse[0]*xyz[0]+inverse[4]*xyz[1]+inverse[8]*xyz[2]);
  inverse[13]=-(inverse[1]*xyz[0]+inverse[5]*xyz[1]+inverse[9]*xyz[2]);
  inverse[14]=-(inverse[2]*xyz[0]+inverse[6]*xyz[1]+inverse[10]*xyz[2]);
  inverse[15]=1;
}

void CholeskyDecomposition(double* dest, const double* a, const int rows, const int cols)
{
  // only needs the lower triangle of the dest positive definte matrix
  // this does not zero the upper triangle of the dest matrix
  assert(rows==cols);

  for (int i=0; i<rows; i++) 
  {
    int jpos=0;
    for (int j=0; j<=i; j++, jpos+=rows)
    {
      int t=0;
      double sum=a[i+jpos];
      for (int k=0; k<j; k++, t+=rows)
        sum-=dest[i+t]*dest[j+t];

      if (i==j) //        if (sum<=0.0) a, with rounding errors, is not positive definite.
        dest[i+i*rows]=sqrt(sum);
      else 
        dest[i+j*rows]=sum/dest[j+jpos];//j*rows];
    }
  }
}

void DumpMatrix(const wchar_t* filename, double* a, const int rows, const int cols)
{
  FILE* file=NULL;
  char buffer[100];
  if (!_wfopen_s(&file, filename, L"wb") && file)
  {
    for (int i=0; i<rows; i++)
    {
      for (int j=0; j<cols; j++)
      {
        _gcvt_s(buffer, 100, a[i+j*cols], 10);
        fwrite(buffer, 1, strlen(buffer), file);
        fwrite("\t ", 1, 2, file);
      }
      fwrite("\r\n", 1, 2, file);
    }
    fclose(file);
  }
}