#include "PDUpdate.h"
#include "Config.h"
#include <assert.h>

#define MAX_COUNT           256                // max number of point "seens" to record for averaging colour and normals

DWORD WINAPI PDUpdateThread(LPVOID lpParameter);

PDUpdate::PDUpdate() : noofthreads(0), threads(NULL), alerts(NULL), threadinfo(NULL)
{
  framesize=DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT;
  localminsq=MATCH_DISTANCE*MATCH_DISTANCE;
  SetNoOfThreads(1);
}

PDUpdate::~PDUpdate() 
{
  KillThreads();
  free(threadinfo);
}

void PDUpdate::KillThreads()
{
  if (noofthreads)
  {
    alive=false;
    for (int i=0; i<noofthreads; i++)
      SetEvent(threadinfo[i].alerts[0]);
  
    WaitForMultipleObjects(noofthreads, threads, TRUE, INFINITE);

    for (int i=0; i<noofthreads; i++)
    {
      CloseHandle(threadinfo[i].alerts[0]);
      CloseHandle(threadinfo[i].alerts[1]);
      free(threadinfo[i].addpts);
    }

    free(threads);
    free(alerts);
    threads=NULL;
    alerts=NULL;
    noofthreads=0;
  }
}

void PDUpdate::SetNoOfThreads(const int n)
{
  if (n!=noofthreads)
  {
    KillThreads();
  
    noofthreads=n;
    threads=(HANDLE*) malloc(sizeof(HANDLE)*n);
    alerts=(HANDLE*) malloc(sizeof(HANDLE)*n);

    alive=true;
    threadinfo=(PDTHREADINFO*) realloc(threadinfo, sizeof(PDTHREADINFO)*n);
    for (int i=0; i<n; i++)
    {
      threadinfo[i].alive=&alive;
      threadinfo[i].index=i;
      threadinfo[i].parent=this;

      threadinfo[i].maxadd=0;
      threadinfo[i].addpts=NULL;

      threadinfo[i].alerts[0]=CreateEvent(NULL, TRUE, FALSE, NULL);
      threadinfo[i].alerts[1]=CreateEvent(NULL, TRUE, FALSE, NULL);
      alerts[i]=threadinfo[i].alerts[1];
      threads[i]=CreateThread(NULL, 0, PDUpdateThread, threadinfo+i, 0, NULL);
    }
  }
}

POINTDATA* PDUpdate::UpdatePointCloud(POINTDATA* frame, const float* transform, int* counts, POINTDATA* worlddata, int& noofwldpts, int& maxwldpts, const DWORD iteration, const bool findmatches)
{
  this->transform=transform;
  this->findmatches=findmatches;

  int i, offset=0;
  int ppt=framesize/noofthreads;
  int rem=framesize-(ppt*noofthreads);

  offset=ppt+rem;
  threadinfo[0].pointdata=frame;
  threadinfo[0].noofpts=offset;
  threadinfo[0].iteration=iteration;
  ResetEvent(threadinfo[0].alerts[1]);
  SetEvent(threadinfo[0].alerts[0]);

  for (i=1; i<noofthreads; i++)
  {
    threadinfo[i].pointdata=frame+offset;
    threadinfo[i].noofpts=ppt;
    threadinfo[i].iteration=iteration;
    ResetEvent(threadinfo[i].alerts[1]);
    SetEvent(threadinfo[i].alerts[0]);

    offset+=ppt;
  }

  WaitForMultipleObjects(noofthreads, alerts, TRUE, INFINITE);

  if (counts)
    counts[0]=counts[1]=0;

  for (i=0; i<noofthreads; i++)
  {
    if (threadinfo[i].additions)
    {
      if (noofwldpts+threadinfo[i].additions>maxwldpts)
      {
        maxwldpts+=2*threadinfo[i].additions;
        POINTDATA* tmp=worlddata;
        worlddata=(POINTDATA*) realloc(worlddata, sizeof(POINTDATA)*maxwldpts);
        if (tmp!=worlddata)
        {
          clouddata->ClearPointData();
          clouddata->AddPointData(worlddata, noofwldpts);
        }
      }
      for (int j=0; j<threadinfo[i].additions; j++)
      {
        worlddata[noofwldpts]=threadinfo[i].addpts[j];
        worlddata[noofwldpts].paired=false;
        worlddata[noofwldpts].used=true;
        worlddata[noofwldpts].lastseen=iteration;
        worlddata[noofwldpts].count=1;

        clouddata->AddPointData(worlddata+noofwldpts);
        ++noofwldpts;
      }

      if (counts)
        counts[0]+=threadinfo[i].additions;
    }
    if (counts)
      counts[1]+=threadinfo[i].updates;
  }

  assert(clouddata->CheckConsistency());

  return worlddata;
}

void PDUpdate::UpdatePointCloudAux(PDTHREADINFO* info)
{
  info->additions=0;
  info->updates=0;

  POINTDATA* pt=info->pointdata;
  for (int i=0; i<info->noofpts; i++, ++pt)
  {
    if (pt->used)
    {
      float txyz[3], nxyz[3];
      POINTDATA* matchpt=NULL;

      txyz[0]=transform[0]*pt->xyz[0]+transform[4]*pt->xyz[1]+transform[8]*pt->xyz[2]+transform[12];
      txyz[1]=transform[1]*pt->xyz[0]+transform[5]*pt->xyz[1]+transform[9]*pt->xyz[2]+transform[13];
      txyz[2]=transform[2]*pt->xyz[0]+transform[6]*pt->xyz[1]+transform[10]*pt->xyz[2]+transform[14];

      nxyz[0]=transform[0]*pt->nxyz[0]+transform[4]*pt->nxyz[1]+transform[8]*pt->nxyz[2];
      nxyz[1]=transform[1]*pt->nxyz[0]+transform[5]*pt->nxyz[1]+transform[9]*pt->nxyz[2];
      nxyz[2]=transform[2]*pt->nxyz[0]+transform[6]*pt->nxyz[1]+transform[10]*pt->nxyz[2];

      if (findmatches)
      {
        int noofpts;
        POINTDATA** tmppoints=clouddata->GetPoints(txyz[0], txyz[1], txyz[2], MATCH_DISTANCE, noofpts, info->index);

        float mindistsq=localminsq;
        for (int j=0; j<noofpts; j++)
        {
          float dx=txyz[0]-tmppoints[j]->xyz[0];
          float dy=txyz[1]-tmppoints[j]->xyz[1];
          float dz=txyz[2]-tmppoints[j]->xyz[2];

          //if (dx>-matchdistance && dx<matchdistance && dy>-matchdistance && dy<matchdistance && dz>-matchdistance && dz<matchdistance)
          {
            float distsq=dx*dx+dy*dy+dz*dz;
            if (distsq<mindistsq)
            {
              matchpt=tmppoints[j];
              mindistsq=distsq;
            }
          }
        }
      }
      else if (pt->matchdist<MATCH_DISTANCE)
        matchpt=pt->matchpoint;

      if (matchpt)
      {
        int tc=matchpt->count+1;

        float tmp=0;
        for (int j=0; j<3; j++)
        {
          // if the position is updated then need to update the clouddata
//          matchpt->xyz[j]=(float) ((txyz[j]+matchpt->xyz[j]*matchpt->count)/tc);
          matchpt->nxyz[j]=(float) ((nxyz[j]+matchpt->nxyz[j]*matchpt->count)/tc);
          matchpt->colour[j]=(float) ((pt->colour[j]+matchpt->colour[j]*matchpt->count)/tc);
          tmp+=matchpt->nxyz[j]*matchpt->nxyz[j];
        }
        if (tmp>0)
        {
          tmp=sqrt(tmp);
          matchpt->nxyz[0]/=tmp;
          matchpt->nxyz[1]/=tmp;
          matchpt->nxyz[2]/=tmp;
        }
        else
        {
          matchpt->nxyz[0]=0.57735f;
          matchpt->nxyz[1]=0.57735f;
          matchpt->nxyz[2]=0.57735f;
        }

        matchpt->lastseen=info->iteration;
        if (tc<MAX_COUNT)
          ++matchpt->count;
        ++info->updates;
      }
      else
      {
        if (info->additions==info->maxadd)
        {
          info->maxadd+=500;
          info->addpts=(POINTDATA*) realloc(info->addpts, sizeof(POINTDATA)*info->maxadd);
        }
        memset(info->addpts+info->additions, 0, sizeof(POINTDATA));
        for (int j=0; j<3; j++)
        {
          info->addpts[info->additions].xyz[j]=txyz[j];
          info->addpts[info->additions].nxyz[j]=nxyz[j];
          info->addpts[info->additions].colour[j]=pt->colour[j];
        }
        info->additions++;
      }
    }
  }
}

DWORD WINAPI PDUpdateThread(LPVOID lpParameter)
{
  PDTHREADINFO* info=(PDTHREADINFO*) lpParameter;
  PDUpdate* parent=(PDUpdate*) info->parent;

  while (*info->alive)
  {
    WaitForSingleObject(info->alerts[0], INFINITE);
    ResetEvent(info->alerts[0]);

    if (*info->alive)
      parent->UpdatePointCloudAux(info);

    SetEvent(info->alerts[1]);
  }

  return 0;
}