#include "PairMatch.h"
#include "Config.h"

DWORD WINAPI PairMatchThread(LPVOID lpParameter);

PairMatch::PairMatch() : noofthreads(0), threads(NULL), threadinfo(NULL), alerts(NULL)
{
  fwh=DEPTH_FRAME_WIDTH/2;
  fhh=DEPTH_FRAME_HEIGHT/2;

  SetNoOfThreads(1);
}

PairMatch::~PairMatch() 
{
  KillThreads();
  free(threadinfo);
}

void PairMatch::KillThreads()
{
  if (noofthreads)
  {
    alive=false;
    for (int i=0; i<noofthreads; i++)
      SetEvent(threadinfo[i].alerts[0]);
  
    WaitForMultipleObjects(noofthreads, threads, TRUE, INFINITE);

    for (int i=0; i<noofthreads; i++)
    {
      CloseHandle(threadinfo[i].alerts[0]);
      CloseHandle(threadinfo[i].alerts[1]);
    }

    free(threads);
    free(alerts);
    threads=NULL;
    alerts=NULL;
    noofthreads=0;
  }
}

void PairMatch::SetNoOfThreads(const int n)
{
  if (n!=noofthreads)
  {
    KillThreads();

    alive=true;
    noofthreads=n;
    threads=(HANDLE*) malloc(sizeof(HANDLE)*n);
    alerts=(HANDLE*) malloc(sizeof(HANDLE)*n);
    threadinfo=(PAIRTHREADINFO*) realloc(threadinfo, sizeof(PAIRTHREADINFO)*n);
    for (int i=0; i<n; i++)
    {
      threadinfo[i].alive=&alive;
      threadinfo[i].parent=this;
      threadinfo[i].threadindex=i;
      threadinfo[i].alerts[0]=CreateEvent(NULL, TRUE, FALSE, NULL);
      threadinfo[i].alerts[1]=CreateEvent(NULL, TRUE, FALSE, NULL);
      alerts[i]=threadinfo[i].alerts[1];
      threads[i]=CreateThread(NULL, 0, PairMatchThread, threadinfo+i, 0, NULL);
    }
  }
}

void PairMatch::FindInversePairs(POINTDATA** invpts, const int noofinvpts, POINTDATA* frame, const double* inverse, const bool usenormal)
{
  this->inverse=inverse;
  this->usenormal=usenormal;

  int i, offset=0;
  int ppt=noofinvpts/noofthreads;
  int rem=noofinvpts-(ppt*noofthreads);

  for (i=0; i<DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT; i++)
    frame[i].matchpoint=NULL;
/*  
  for (i=0; i<noofinvpts; i++)
  {
    invpts[i]->matchpoint=NULL;
    invpts[i]->matchdist=DISTANCE_THRESHOLD*DISTANCE_THRESHOLD;
  }*/

  offset=ppt+rem;
  threadinfo[0].invpts=invpts;
  threadinfo[0].noofinvpts=offset;
  threadinfo[0].frame=frame;
  threadinfo[0].version=3;
  ResetEvent(threadinfo[0].alerts[1]);
  SetEvent(threadinfo[0].alerts[0]);

  for (i=1; i<noofthreads; i++)
  {
    threadinfo[i].invpts=invpts+offset;
    threadinfo[i].noofinvpts=ppt;
    threadinfo[i].frame=frame;
    threadinfo[i].version=3;
    ResetEvent(threadinfo[i].alerts[1]);
    SetEvent(threadinfo[i].alerts[0]);

    offset+=ppt;
  }

  WaitForMultipleObjects(noofthreads, alerts, TRUE, INFINITE);
}

int PairMatch::FindInversePairsAux(PAIRTHREADINFO* info)
{
  const float THRESHOLD=MATCH_DISTANCE*MATCH_DISTANCE;

  for (int i=0; i<info->noofinvpts; i++)
  {
    POINTDATA* wpt=info->invpts[i];
    wpt->matchpoint=NULL;
    wpt->matchdist=THRESHOLD;

    float wpx=(float) (inverse[0]*wpt->xyz[0]+inverse[4]*wpt->xyz[1]+inverse[8]*wpt->xyz[2]+inverse[12]);
    float wpy=(float) (inverse[1]*wpt->xyz[0]+inverse[5]*wpt->xyz[1]+inverse[9]*wpt->xyz[2]+inverse[13]);
    float wpz=(float) (inverse[2]*wpt->xyz[0]+inverse[6]*wpt->xyz[1]+inverse[10]*wpt->xyz[2]+inverse[14]);

    float wpnx=(float) (inverse[0]*wpt->nxyz[0]+inverse[4]*wpt->nxyz[1]+inverse[8]*wpt->nxyz[2]);
    float wpny=(float) (inverse[1]*wpt->nxyz[0]+inverse[5]*wpt->nxyz[1]+inverse[9]*wpt->nxyz[2]);
    float wpnz=(float) (inverse[2]*wpt->nxyz[0]+inverse[6]*wpt->nxyz[1]+inverse[10]*wpt->nxyz[2]);

    float depth=-wpz*1000;

    float scale=depth*PERSPECTIVE_CORRECTION*WORLD_SCALE;
    int tpx=(int) (wpx/scale);
    int tpy=(int) (wpy/scale);

    int DIFF=(int) (MATCH_DISTANCE/(depth*PERSPECTIVE_CORRECTION*WORLD_SCALE));
    if (DIFF<=0)
      DIFF=1;

    int si=fwh+tpx-DIFF;
    int sj=fhh+tpy-DIFF;
      
    if (si<0)
      si=0;
    else if (si>=DEPTH_FRAME_WIDTH)
      si=DEPTH_FRAME_WIDTH-1;

    if (sj<0)
      sj=0;
    else if (sj>=DEPTH_FRAME_HEIGHT)
      sj=DEPTH_FRAME_HEIGHT-1;

    int ei=si+(int) (2*DIFF);
    int ej=sj+(int) (2*DIFF);
    if (ei>=DEPTH_FRAME_WIDTH)
      ei=DEPTH_FRAME_WIDTH-1;
    if (ej>=DEPTH_FRAME_HEIGHT)
      ej=DEPTH_FRAME_HEIGHT-1;

    int offset=si+sj*DEPTH_FRAME_WIDTH;
    for (int y=sj; y<=ej; y++, offset+=DEPTH_FRAME_WIDTH)
    {
      POINTDATA* pt=info->frame+offset;
      for (int x=si; x<=ei; x++, ++pt)
      {
        if (pt->used)
        {
          float dz=pt->xyz[2]-wpz;

          //if (dz>-DISTANCE_THRESHOLD && dz<DISTANCE_THRESHOLD)
          {
            float dx=pt->xyz[0]-wpx;
            float dy=pt->xyz[1]-wpy;
            float distsq=(dx*dx+dy*dy+dz*dz);
            if (distsq<THRESHOLD)
            {
              pt->matchpoint=wpt;

              if (distsq<wpt->matchdist)
              {
                float tmp=(usenormal?pt->nxyz[0]*wpnx+pt->nxyz[1]*wpny+pt->nxyz[2]*wpnz:1.0f);
                if (tmp>INV_NORMAL_THRESHOLD)
                {
                  wpt->matchpoint=pt;
                  wpt->matchdist=distsq;
                }
              }
            }
          }
        }
      }
    }
  }

  return 0;
}

void PairMatch::FindPairsV1(POINTDATA** invpts, const int noofinvpts, POINTDATA* frame, const double* inverse, const bool usenormal)
{
  this->inverse=inverse;
  this->usenormal=usenormal;

  int i, offset=0;
  int ppt=noofinvpts/noofthreads;
  int rem=noofinvpts-(ppt*noofthreads);
  
  for (i=0; i<DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT; i++)
  {
    frame[i].matchpoint=NULL;
    frame[i].matchdist=DISTANCE_THRESHOLD*DISTANCE_THRESHOLD;
  }

  offset=ppt+rem;
  threadinfo[0].invpts=invpts;
  threadinfo[0].noofinvpts=offset;
  threadinfo[0].frame=frame;
  threadinfo[0].version=1;
  ResetEvent(threadinfo[0].alerts[1]);
  SetEvent(threadinfo[0].alerts[0]);

  for (i=1; i<noofthreads; i++)
  {
    threadinfo[i].invpts=invpts+offset;
    threadinfo[i].noofinvpts=ppt;
    threadinfo[i].frame=frame;
    threadinfo[i].version=1;
    ResetEvent(threadinfo[i].alerts[1]);
    SetEvent(threadinfo[i].alerts[0]);

    offset+=ppt;
  }

  WaitForMultipleObjects(noofthreads, alerts, TRUE, INFINITE);
}

int PairMatch::FindPairsV1Aux(PAIRTHREADINFO* info)
{
  const float range=DISTANCE_THRESHOLD;
  const int diff=(int) (2*range/WORLD_SCALE);

  for (int i=0; i<info->noofinvpts; i++)
  {
    POINTDATA* wpt=info->invpts[i];
//    wpt->paired=false;

    float wpx=(float) (inverse[0]*wpt->xyz[0]+inverse[4]*wpt->xyz[1]+inverse[8]*wpt->xyz[2]+inverse[12]);
    float wpy=(float) (inverse[1]*wpt->xyz[0]+inverse[5]*wpt->xyz[1]+inverse[9]*wpt->xyz[2]+inverse[13]);
    float wpz=(float) (inverse[2]*wpt->xyz[0]+inverse[6]*wpt->xyz[1]+inverse[10]*wpt->xyz[2]+inverse[14]);

    float wpnx=(float) (inverse[0]*wpt->nxyz[0]+inverse[4]*wpt->nxyz[1]+inverse[8]*wpt->nxyz[2]);
    float wpny=(float) (inverse[1]*wpt->nxyz[0]+inverse[5]*wpt->nxyz[1]+inverse[9]*wpt->nxyz[2]);
    float wpnz=(float) (inverse[2]*wpt->nxyz[0]+inverse[6]*wpt->nxyz[1]+inverse[10]*wpt->nxyz[2]);

    int si=fwh+(int) ((wpx-range)/WORLD_SCALE);
    int sj=fhh+(int) ((wpy-range)/WORLD_SCALE);
      
    if (si<0)
      si=0;
    else if (si>=DEPTH_FRAME_WIDTH)
      si=DEPTH_FRAME_WIDTH-1;

    if (sj<0)
      sj=0;
    else if (sj>=DEPTH_FRAME_HEIGHT)
      sj=DEPTH_FRAME_HEIGHT-1;

    int ei=si+diff;
    int ej=sj+diff;
    if (ei>=DEPTH_FRAME_WIDTH)
      ei=DEPTH_FRAME_WIDTH-1;
    if (ej>=DEPTH_FRAME_HEIGHT)
      ej=DEPTH_FRAME_HEIGHT-1;

    POINTDATA* localpt=NULL;
    float localminsq=DISTANCE_THRESHOLD*DISTANCE_THRESHOLD;//*LOCAL_MIN;
    float matchweight;
    int offset=si+sj*DEPTH_FRAME_WIDTH;
    for (int y=sj; y<=ej; y++, offset+=DEPTH_FRAME_WIDTH)
    {
      POINTDATA* pt=info->frame+offset;
      for (int x=si; x<=ei; x++, ++pt)
      {
        if (pt->used)
        {
          float dz=pt->xyz[2]-wpz;
//          dz/=20.0f;

          //if (dz>-DISTANCE_THRESHOLD && dz<DISTANCE_THRESHOLD)
          {
            float dx=pt->xyz[0]-wpx;
            float dy=pt->xyz[1]-wpy;
            float distsq=(dx*dx+dy*dy+dz*dz);
            if (distsq<localminsq)
            {
              float tmp=(usenormal?pt->nxyz[0]*wpnx+pt->nxyz[1]*wpny+pt->nxyz[2]*wpnz:1.0f);
              if (tmp>INV_NORMAL_THRESHOLD)
              {
                localpt=pt;
                localminsq=distsq;
                matchweight=tmp;
              }
            }
          }
        }
      }
    }

    float localmin=sqrt(localminsq);
    if (localpt && localmin<localpt->matchdist)
    {
      localpt->matchpoint=info->invpts[i];
      localpt->matchdist=localmin;
      localpt->matchweight=CalcMatchWeight(matchweight);
    }
  }

  return 0;//noofpairs;
}

void PairMatch::FindPairsV2(POINTDATA* frame, const double* transform, CloudData* clouddata, const bool usenormal)
{
  this->transform=transform;
  this->usenormal=usenormal;

  const int framesize=DEPTH_FRAME_WIDTH*DEPTH_FRAME_HEIGHT;
  int i, offset=0;
  int ppt=framesize/noofthreads;
  int rem=framesize-(ppt*noofthreads);

  offset=ppt+rem;
  threadinfo[0].frame=frame;
  threadinfo[0].noofpts=offset;
  threadinfo[0].clouddata=clouddata;
  threadinfo[0].version=2;
  ResetEvent(threadinfo[0].alerts[1]);
  SetEvent(threadinfo[0].alerts[0]);

  for (i=1; i<noofthreads; i++)
  {
    threadinfo[i].frame=frame+offset;
    threadinfo[i].noofpts=ppt;
    threadinfo[i].clouddata=clouddata;
    threadinfo[i].version=2;
    ResetEvent(threadinfo[i].alerts[1]);
    SetEvent(threadinfo[i].alerts[0]);
    offset+=ppt;
  }

  WaitForMultipleObjects(noofthreads, alerts, TRUE, INFINITE);
}

int PairMatch::FindPairsV2Aux(PAIRTHREADINFO* info)
{
  CloudData* clouddata=info->clouddata;

  POINTDATA* pt=info->frame;
  for (int i=0; i<info->noofpts; i++, ++pt)
  {
    if (pt->used)
    {
      float txyz[3], nxyz[3];
      txyz[0]=(float) (transform[0]*pt->xyz[0]+transform[4]*pt->xyz[1]+transform[8]*pt->xyz[2]+transform[12]);
      txyz[1]=(float) (transform[1]*pt->xyz[0]+transform[5]*pt->xyz[1]+transform[9]*pt->xyz[2]+transform[13]);
      txyz[2]=(float) (transform[2]*pt->xyz[0]+transform[6]*pt->xyz[1]+transform[10]*pt->xyz[2]+transform[14]);

      nxyz[0]=(float) (transform[0]*pt->nxyz[0]+transform[4]*pt->nxyz[1]+transform[8]*pt->nxyz[2]);
      nxyz[1]=(float) (transform[1]*pt->nxyz[0]+transform[5]*pt->nxyz[1]+transform[9]*pt->nxyz[2]);
      nxyz[2]=(float) (transform[2]*pt->nxyz[0]+transform[6]*pt->nxyz[1]+transform[10]*pt->nxyz[2]);

      int noofpts;
      POINTDATA** tmppoints=clouddata->GetPoints(txyz[0], txyz[1], txyz[2], DISTANCE_THRESHOLD, noofpts, info->threadindex);
      
      float mindistsqr=DISTANCE_THRESHOLD*DISTANCE_THRESHOLD;
      float matchweight;

      POINTDATA* matchpt=NULL;
      for (int j=0; j<noofpts; j++)
      {
        POINTDATA* wpt=tmppoints[j];
        float dx=txyz[0]-wpt->xyz[0];
        float dy=txyz[1]-wpt->xyz[1];
        float dz=txyz[2]-wpt->xyz[2];
        float distsqr=(dx*dx+dy*dy+dz*dz);
        if (distsqr<mindistsqr)
        {
          float tmp=(usenormal?nxyz[0]*wpt->nxyz[0]+nxyz[1]*wpt->nxyz[1]+nxyz[2]*wpt->nxyz[2]:1.0f);
          if (tmp>INV_NORMAL_THRESHOLD)
          {
            mindistsqr=distsqr;
            matchpt=wpt;
            matchweight=tmp;
          }
        }
      }

      if (pt->matchpoint=matchpt)
      {
        pt->matchdist=(float) sqrt(mindistsqr);
        pt->matchweight=CalcMatchWeight(matchweight);
      }
    }
    else
      pt->matchpoint=NULL;
  }

  return 0;
}

DWORD WINAPI PairMatchThread(LPVOID lpParameter)
{
  PAIRTHREADINFO* info=(PAIRTHREADINFO*) lpParameter;
  PairMatch* parent=(PairMatch*) info->parent;

  while (*info->alive)
  {
    WaitForSingleObject(info->alerts[0], INFINITE);
    ResetEvent(info->alerts[0]);

    if (*info->alive)
    {
      if (info->version==1)
        parent->FindPairsV1Aux(info);
      else if (info->version==2)
        parent->FindPairsV2Aux(info);
      else if (info->version==3)
        parent->FindInversePairsAux(info);
    }

    SetEvent(info->alerts[1]);
  }

  return 0;
}
