#pragma once
#include "CloudData.h"
#include "Config.h"
#include "GLRenderer.h"
#include <math.h>

struct PAIRTHREADINFO
{
  VOID* parent;

  bool* alive;
  HANDLE alerts[2];   // 0=start; 1=end

  int version;
  int noofinvpts;
  POINTDATA** invpts, *frame;
  
  // extra info used for version 2
  int noofpts, threadindex;
  CloudData* clouddata;
};

class PairMatch
{
public:
  PairMatch();
  ~PairMatch();

  void KillThreads();
  void SetNoOfThreads(const int n);
  void SetFrameSize(const int width, const int height);

  void FindPairsV1(POINTDATA** invpts, const int noofinvpts, POINTDATA* frame, const double* inverse, const bool usenormal);
  void FindPairsV2(POINTDATA* frame, const double* transform, CloudData* clouddata, const bool usenormal);
  void FindInversePairs(POINTDATA** invpts, const int noofinvpts, POINTDATA* frame, const double* inverse, const bool usenormal);

private:
  int noofthreads;
  bool alive;
  HANDLE* threads, *alerts;
  PAIRTHREADINFO* threadinfo;

  int fwh, fhh;//, diff;
//  float range;

  bool usenormal;
  const double* inverse, *transform;

  float CalcMatchWeight(float weight);

  int FindPairsV1Aux(PAIRTHREADINFO* info);
  int FindPairsV2Aux(PAIRTHREADINFO* info);
  int FindInversePairsAux(PAIRTHREADINFO* info);

  friend DWORD WINAPI PairMatchThread(LPVOID lpParameter);
};

inline float PairMatch::CalcMatchWeight(float weight)
{
  const float RANGE=1.0f/(1.0f-INV_NORMAL_THRESHOLD);
  return (weight-INV_NORMAL_THRESHOLD)*RANGE;
}
