#include "MainWnd.h"

void MainWnd::TestCodeV2()
{
  const int noofpts=7;
  const double x[noofpts]={0.038, 0.194, 0.425, 0.626, 1.253, 2.500, 3.740};
  const double y[noofpts]={0.050, 0.127, 0.094, 0.2122, 0.2729, 0.2665, 0.3317};

  double r[noofpts], update[2], b[2], yt[2];
  double jacobian[noofpts*2], lowermat[4], tmpsqr[4];
//  double t1[2][2], inv[2][2], t2[2][noofpts];

  double v=0.9;
  double k=0.2;

  int loop=1;
  do
  {
    double error=0;
    for (int i=0; i<noofpts; i++)
    {
      r[i]=y[i]-((v*x[i])/(k+x[i]));
      error+=r[i]*r[i];
    }

    for (int i=0; i<noofpts; i++)
    {
      double t=k+x[i];
      jacobian[i]=x[i]/t;
      jacobian[i+noofpts]=-((v*x[i])/(t*t));
    }

    MatMultTA_LT(tmpsqr, jacobian, noofpts, 2, jacobian, noofpts, 2);
    CholeskyDecomposition(lowermat, tmpsqr, 2, 2);

    MatMultTA(b, jacobian, noofpts, 2, r, noofpts);
    for (int i=0; i<2; i++)
    {
      int ti=0;
      double sum=b[i];
      for (int j=0; j<i; j++, ti+=2)
        sum-=lowermat[i+ti]*yt[j];
      yt[i]=sum/lowermat[i+ti];
    }

    for (int i=1; i>=0; --i)
    {
      double sum=yt[i];
      int ti=i*2;
      for (int j=i+1; j<2; j++)
        sum-=lowermat[j+ti]*update[j];
      update[i]=sum/lowermat[ti+i];
    }

    double tmperror=0;
    double alpha=1.0;
/*
    double targeterror=0;
    for (int i=0; i<noofpts; i+=3)
      targeterror+=r[i]*r[i];

    do
    {
      double tv=v+alpha*update[0];
      double tk=k+alpha*update[1];

      tmperror=0;
      for (int i=0; i<noofpts; i+=3)
      {
        double t=y[i]-((tv*x[i])/(tk+x[i]));
        tmperror+=t*t;
      }

      if (tmperror>targeterror)
        alpha/=2;
    } while (tmperror>targeterror && alpha>0.02);
    */

    v+=alpha*update[0];
    k+=alpha*update[1];

    ++loop;
  
  } while (loop<100);

}

void MainWnd::TestCode()
{
  const int noofpts=7;
  const double x[noofpts]={0.038, 0.194, 0.425, 0.626, 1.253, 2.500, 3.740};
  const double y[noofpts]={0.050, 0.127, 0.094, 0.2122, 0.2729, 0.2665, 0.3317};

  double r[noofpts], update[2];
  double jacobian[noofpts][2];
  double t1[2][2], inv[2][2], t2[2][noofpts];

  double v=0.9;
  double k=0.2;

  int loop=1;
  do
  {
    double error=0;
    for (int i=0; i<noofpts; i++)
    {
      r[i]=y[i]-((v*x[i])/(k+x[i]));
      error+=r[i]*r[i];
    }

    for (int i=0; i<noofpts; i++)
    {
      double t=k+x[i];
      jacobian[i][0]=x[i]/t;
      jacobian[i][1]=-((v*x[i])/(t*t));
    }

    for (int i=0; i<2; i++)
    {
      for (int j=0; j<2; j++)
      {
        t1[i][j]=0;
        for (int k=0; k<noofpts; k++)
        {
          t1[i][j]+=jacobian[k][i]*jacobian[k][j];
        }
      }
    }

    double det=1/((t1[0][0]*t1[1][1])-(t1[0][1]*t1[1][0]));
    inv[0][0]=t1[1][1]*det;
    inv[0][1]=-t1[0][0]*det;
    inv[1][0]=-t1[1][0]*det;
    inv[1][1]=t1[0][0]*det;

    double tst1=t1[0][0]*inv[0][0]+t1[0][1]*inv[1][0];
    double tst2=t1[1][0]*inv[0][0]+t1[1][1]*inv[1][0];
    double tst3=t1[0][0]*inv[1][0]+t1[0][1]*inv[1][1];
    double tst4=t1[1][0]*inv[1][0]+t1[1][1]*inv[1][1];

    for (int i=0; i<2; i++)
    {
      for (int j=0; j<noofpts; j++)
      {
        t2[i][j]=0;
        for (int k=0; k<2; k++)
        {
          t2[i][j]+=inv[i][k]*jacobian[j][k];
        }
      }
    }

    for (int i=0; i<2; i++)
    {
      update[i]=0;
      for (int k=0; k<noofpts; k++)
        update[i]+=t2[i][k]*r[k];
    }

/*    while (update[0]<-0.1 || update[0]>0.1 || update[1]<-0.1 || update[1]>0.1)
    {
      update[0]/=2;
      update[1]/=2;
    }*/

    double tmperror=0;
    double alpha=1.0;

    double targeterror=0;
    for (int i=0; i<noofpts; i+=3)
      targeterror+=r[i]*r[i];

    do
    {
      double tv=v+alpha*update[0];
      double tk=k+alpha*update[1];

      tmperror=0;
      for (int i=0; i<noofpts; i+=3)
      {
        double t=y[i]-((tv*x[i])/(tk+x[i]));
        tmperror+=t*t;
      }

      if (tmperror>targeterror)
        alpha/=2;
    } while (tmperror>targeterror && alpha>0.02);


    v+=alpha*update[0];
    k+=alpha*update[1];
//    v=update[0];
//    k=update[1];

    ++loop;
  
  } while (loop<100);

}